"""Forcing (:mod:`fluidsim.solvers.ns2d.strat.forcing`)
=======================================================

.. autoclass:: ForcingNS2DStrat
   :members:

"""

from fluidsim.solvers.ns2d.forcing import ForcingNS2D


class ForcingNS2DStrat(ForcingNS2D):

    """Forcing class for the ns2d strat solver.

    .. inheritance-diagram:: ForcingNS2DStrat

    """

    def compute_coef_ab_normalize(
        self, constant_rate_of, key_forced, f_fft, var_fft, deltat
    ):

        if constant_rate_of not in ["energy", "energyK"]:
            raise ValueError

        if hasattr(self.forcing_maker, "oper_coarse"):
            oper = self.forcing_maker.oper_coarse
            state = self.forcing_maker.fstate_coarse
        else:
            oper = self.sim.oper
            state = self.sim.state

        sum_k = oper.sum_wavenumbers

        if key_forced == "rot_fft":
            vx_fft, vy_fft = oper.vecfft_from_rotfft(var_fft)
            fx_fft, fy_fft = oper.vecfft_from_rotfft(f_fft)
            a = deltat / 2 * sum_k(abs(fx_fft) ** 2 + abs(fy_fft) ** 2)
            b = sum_k(
                (vx_fft.conj() * fx_fft).real + (vy_fft.conj() * fy_fft).real
            )

            return a, b
        elif key_forced == "ap_fft":
            rot_fft, b_fft = state.compute_rotbfft_from_apfft(var_fft)
            frot_fft, fb_fft = state.compute_rotbfft_from_apfft(f_fft)

            vx_fft, vy_fft = oper.vecfft_from_rotfft(rot_fft)
            fx_fft, fy_fft = oper.vecfft_from_rotfft(frot_fft)

        else:
            raise ValueError

        a = deltat / 2 * sum_k(abs(fx_fft) ** 2 + abs(fy_fft) ** 2)
        b = sum_k((vx_fft.conj() * fx_fft).real + (vy_fft.conj() * fy_fft).real)

        if constant_rate_of == "energyK":
            return a, b

        N2 = self.sim.params.N ** 2
        a += deltat / 2 * sum_k(abs(fb_fft) ** 2) / N2
        b += sum_k((b_fft.conj() * fb_fft).real) / N2

        return a, b



import unittest
import numpy as np
import sys
from copy import deepcopy

import fluiddyn.util.mpi as mpi

from fluidsim.util.testing import TestCase, skip_if_no_fluidfft


def create_oper(type_fft=None, coef_dealiasing=2.0 / 3, **kwargs):
    from fluidsim.operators.operators2d import OperatorsPseudoSpectral2D

    params = OperatorsPseudoSpectral2D._create_default_params()

    params.ONLY_COARSE_OPER = kwargs.get("ONLY_COARSE_OPER", False)

    params._set_attrib("f", 0)
    params._set_attrib("c2", 100)
    params._set_attrib("kd2", 0)

    if "nh" in kwargs:
        nh = kwargs["nh"]
    elif mpi.nb_proc == 1:
        nh = 9
    else:
        nh = 8

    params.oper.nx = nh
    params.oper.ny = nh
    Lh = 6.0
    params.oper.Lx = Lh
    params.oper.Ly = Lh

    if type_fft is not None:
        params.oper.type_fft = type_fft

    params.oper.coef_dealiasing = coef_dealiasing

    oper = OperatorsPseudoSpectral2D(params=params)

    return oper


def compute_increments_dim1_old(var, irx):
    """Old version of the function compute_increments_dim1."""
    n0 = var.shape[0]
    n1 = var.shape[1]
    n1new = n1 - irx
    inc_var = np.empty([n0, n1new])
    for i0 in range(n0):
        for i1 in range(n1new):
            inc_var[i0, i1] = var[i0, i1 + irx] - var[i0, i1]
    return inc_var


@skip_if_no_fluidfft
@unittest.skipIf(sys.platform.startswith("win"), "Untested on Windows")
class TestOperators(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.oper = create_oper()
        cls.rtol = 1e-15
        cls.atol = 1e-14  # Absolute tolerance for double precision FFT

    def test_curl(self):
        """Test curl"""
        oper = self.oper
        rot = oper.create_arrayX_random()
        rot_fft = oper.fft2(rot)
        if mpi.rank == 0:
            rot_fft[0, 0] = 0.0

        ux_fft, uy_fft = oper.vecfft_from_rotfft(rot_fft)
        rot2_fft = oper.rotfft_from_vecfft(ux_fft, uy_fft)

        np.testing.assert_allclose(rot2_fft, rot_fft, self.rtol, self.atol)

    def test_laplacian(self):
        oper = self.oper
        ff = oper.create_arrayX_random()
        ff_fft = oper.fft2(ff)
        if mpi.rank == 0:
            ff_fft[0, 0] = 0.0

        lap_fft = oper.laplacian_fft(ff_fft, order=4)
        ff_fft_back = oper.invlaplacian_fft(lap_fft, order=4)

        np.testing.assert_allclose(ff_fft, ff_fft_back, self.rtol, self.atol)

        lap_fft = oper.laplacian_fft(ff_fft, order=2, negative=True)
        invlap_fft = oper.invlaplacian_fft(ff_fft, order=2, negative=True)

        np.testing.assert_equal(oper.K2 * ff_fft, lap_fft)
        np.testing.assert_allclose(
            ff_fft / oper.K2_not0, invlap_fft, self.rtol, self.atol
        )

    def test_compute_increments_dim1(self):
        """Test computing increments of var over the dim 1."""
        oper = self.oper
        var = oper.create_arrayX_random()

        def assert_increments_equal(irx):
            inc_var = oper.compute_increments_dim1(var, irx)
            inc_var_old = compute_increments_dim1_old(var, irx)
            np.testing.assert_equal(inc_var_old, inc_var)

        n1 = oper.shapeX[1]
        for irx in [n1, n1 // 2, 0]:
            assert_increments_equal(irx)


@skip_if_no_fluidfft
class TestOperatorOnlyCoarse(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.nh = 12
        cls.oper = create_oper(ONLY_COARSE_OPER=True, nh=cls.nh)

    def test_oper_coarse(self):
        """Test coarse operator parameters which, by default, initializes
        `nh=4`.

        """
        oper = self.oper

        # Assert params are intact but the operator is initialized coarse
        self.assertEqual(oper.params.oper.nx, self.nh)
        self.assertEqual(oper.params.oper.ny, self.nh)
        self.assertNotEqual(oper.nx, self.nh)
        self.assertNotEqual(oper.ny, self.nh)


@skip_if_no_fluidfft
class TestOperatorsDealiasing(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.oper = create_oper(coef_dealiasing=False)

    def test_dealiasing(self):
        """Test if dealiasing with coef_dealiasing=1.0 keeps the original
        array unchanged.

        """
        oper = self.oper
        var_fft = oper.create_arrayK(1.0)
        sum_var = var_fft.sum()
        oper.dealiasing(var_fft)
        sum_var_dealiased = var_fft.sum()
        self.assertEqual(sum_var, sum_var_dealiased)


@skip_if_no_fluidfft
class TestCoarse(unittest.TestCase):
    nb_dim = 2

    @property
    def Oper(self):
        from fluidsim.operators.operators2d import OperatorsPseudoSpectral2D

        return OperatorsPseudoSpectral2D

    def test_coarse(self):

        params = self.Oper._create_default_params()
        params.oper.nx = 32
        params.oper.ny = 48
        if self.nb_dim == 3:
            params.oper.nz = 12

        params.oper.truncation_shape = "spherical"

        oper = self.Oper(params)

        params_coarse = deepcopy(params)
        params_coarse.oper.nx = 2 * 8
        params_coarse.oper.ny = 2 * 12
        if self.nb_dim == 3:
            params_coarse.oper.nz = 2 * 4

        params_coarse.oper.type_fft = "sequential"
        params_coarse.oper.coef_dealiasing = 1.0

        if mpi.rank == 0:
            oper_coarse = oper.__class__(params=params_coarse)
            oper_coarse_shapeK_loc = oper_coarse.shapeK_loc

            field_coarse = oper_coarse.create_arrayX_random()
            field_coarse_fft = oper_coarse.fft(field_coarse)

            if self.nb_dim == 2:

                nkyc, nkxc = oper_coarse_shapeK_loc
                # zeros because of conditions in put_coarse_array_in_array_fft
                field_coarse_fft[nkyc // 2, :] = 0
                field_coarse_fft[:, nkxc - 1] = 0
            elif self.nb_dim == 3:
                nkzc, nkyc, nkxc = oper_coarse_shapeK_loc
                field_coarse_fft[nkzc // 2, :] = 0
                field_coarse_fft[:, nkyc // 2, :] = 0
                field_coarse_fft[:, :, nkxc - 1] = 0

            field_coarse = oper_coarse.ifft(field_coarse_fft)
            energy = oper_coarse.compute_energy_from_X(field_coarse)
        else:
            field_coarse_fft = None
            oper_coarse = None
            oper_coarse_shapeK_loc = None

        if mpi.nb_proc > 1:
            oper_coarse_shapeK_loc = mpi.comm.bcast(
                oper_coarse_shapeK_loc, root=0
            )

        field_fft = oper.create_arrayK(value=0)
        oper.put_coarse_array_in_array_fft(
            field_coarse_fft, field_fft, oper_coarse, oper_coarse_shapeK_loc
        )

        energy_big_fft = oper.compute_energy_from_K(field_fft)

        if mpi.rank == 0:
            print(
                "energy,  energy_big_fft\n"
                + (2 * "{:.8f}    ").format(energy, energy_big_fft)
            )
            assert energy > 0
            assert np.allclose(energy, energy_big_fft)
            print("OK np.allclose(energy, energy_big_fft)")

        field = oper.ifft(field_fft)
        energy_big = oper.compute_energy_from_X(field)

        field_fft_back = oper.fft(field)
        if not np.allclose(field_fft, field_fft_back):
            print("Buggy! field_fft != field_fft_back")

        field_coarse_fft_back = oper.coarse_seq_from_fft_loc(
            field_fft, oper_coarse_shapeK_loc
        )

        if mpi.rank == 0:
            field_coarse_back = oper_coarse.ifft(field_coarse_fft_back)
            energy_back = oper_coarse.compute_energy_from_X(field_coarse_back)
            print(
                "energy,  energy_back,  energy_big_fft,  energy_big\n"
                + (4 * "{:.8f}    ").format(
                    energy, energy_back, energy_big_fft, energy_big
                )
            )
            assert np.allclose(energy, energy_back)
            print("OK np.allclose(energy, energy_back)")
            assert np.allclose(energy, energy_big)


if __name__ == "__main__":
    unittest.main()



"""Modified SW1L equations
==========================

(:mod:`fluidsim.solvers.sw1l.modified.solver`)

This class is a solver of a modified version of the 1 layer shallow
water (Saint Venant) equations for which the advection is only
due to the rotational velocity.
"""

import numpy as np

from fluidsim.base.setofvariables import SetOfVariables

from fluidsim.solvers.sw1l.solver import InfoSolverSW1L
from fluidsim.solvers.sw1l.solver import Simul as SimulSW1L


class InfoSolverSW1LModified(InfoSolverSW1L):
    """Information about the solver SW1L."""

    def _init_root(self, **kargs):
        super()._init_root()

        sw1l = "fluidsim.solvers.sw1l"

        self.module_name = sw1l + ".modified.solver"
        self.class_name = "Simul"
        self.short_name = "SW1Lmodif"

        classes = self.classes

        classes.State.module_name = sw1l + ".modified.state"
        classes.State.class_name = "StateSW1LModified"

        classes.Output.module_name = sw1l + ".modified.output"
        classes.Output.class_name = "OutputSW1LModified"


class Simul(SimulSW1L):
    """A solver of the shallow-water 1 layer equations (SW1L)"""

    InfoSolver = InfoSolverSW1LModified

    def tendencies_nonlin(self, state_spect=None, old=None):
        oper = self.oper
        fft2 = oper.fft2
        ifft2 = oper.ifft2

        if state_spect is None:
            state_phys = self.state.state_phys
            state_spect = self.state.state_spect
        else:
            state_phys = self.state.return_statephys_from_statespect(state_spect)

        ux = state_phys.get_var("ux")
        uy = state_phys.get_var("uy")
        # eta = state_phys.get_var('eta')

        ux_fft = state_spect.get_var("ux_fft")
        uy_fft = state_spect.get_var("uy_fft")
        eta_fft = state_spect.get_var("eta_fft")

        # compute Fx_fft and Fy_fft
        rot_fft = oper.rotfft_from_vecfft(ux_fft, uy_fft)
        ux_rot_fft, uy_rot_fft = oper.vecfft_from_rotfft(rot_fft)
        ux_rot = ifft2(ux_rot_fft)
        uy_rot = ifft2(uy_rot_fft)

        dxux_fft, dyux_fft = oper.gradfft_from_fft(ux_fft)
        dxux = ifft2(dxux_fft)
        dyux = ifft2(dyux_fft)
        dxuy_fft, dyuy_fft = oper.gradfft_from_fft(uy_fft)
        dxuy = ifft2(dxuy_fft)
        dyuy = ifft2(dyuy_fft)

        FNLx = -ux_rot * dxux - uy_rot * dyux
        FNLy = -ux_rot * dxuy - uy_rot * dyuy

        if self.params.beta != 0:
            f = self.params.f + self.params.beta * oper.YY
        else:
            f = self.params.f

        FCx = +f * uy
        FCy = -f * ux

        Fgradx_fft, Fgrady_fft = oper.gradfft_from_fft(self.params.c2 * eta_fft)

        Fx_fft = fft2(FCx + FNLx) - Fgradx_fft
        Fy_fft = fft2(FCy + FNLy) - Fgrady_fft

        # compute Feta_fft
        dxeta_fft, dyeta_fft = oper.gradfft_from_fft(eta_fft)
        dxeta = ifft2(dxeta_fft)
        dyeta = ifft2(dyeta_fft)

        div_fft = oper.divfft_from_vecfft(ux_fft, uy_fft)
        Feta_fft = -fft2(ux_rot * dxeta + uy_rot * dyeta) - div_fft

        if old is None:
            tendencies_fft = SetOfVariables(
                like=self.state.state_spect, info="tendencies_nonlin"
            )
        else:
            tendencies_fft = old

        tendencies_fft.set_var("ux_fft", Fx_fft)
        tendencies_fft.set_var("uy_fft", Fy_fft)
        tendencies_fft.set_var("eta_fft", Feta_fft)

        oper.dealiasing(tendencies_fft)

        if self.params.forcing.enable:
            tendencies_fft += self.forcing.get_forcing()

        return tendencies_fft


if __name__ == "__main__":

    import fluiddyn as fld

    params = Simul.create_default_params()

    params.short_name_type_run = "test"

    nh = 64
    Lh = 2 * np.pi
    params.oper.nx = nh
    params.oper.ny = nh
    params.oper.Lx = Lh
    params.oper.Ly = Lh

    delta_x = params.oper.Lx / params.oper.nx
    params.nu_8 = (
        2.0 * 10e-1 * params.forcing.forcing_rate ** (1.0 / 3) * delta_x ** 8
    )

    params.time_stepping.t_end = 2.0

    params.init_fields.type = "noise"

    params.forcing.enable = True
    params.forcing.type = "waves"

    params.output.periods_print.print_stdout = 0.25

    params.output.periods_save.phys_fields = 1.0
    params.output.periods_save.spectra = 0.5
    params.output.periods_save.spect_energy_budg = 0.5
    params.output.periods_save.increments = 0.5
    params.output.periods_save.pdf = 0.5
    params.output.periods_save.time_signals_fft = False

    params.output.periods_plot.phys_fields = 0.0

    params.output.phys_fields.field_to_plot = "div"

    # params.output.spectra.HAS_TO_PLOT_SAVED = True
    # params.output.spatial_means.HAS_TO_PLOT_SAVED = True
    # params.output.spect_energy_budg.HAS_TO_PLOT_SAVED = True
    # params.output.increments.HAS_TO_PLOT_SAVED = True

    sim = Simul(params)

    # sim.output.phys_fields.plot()
    sim.time_stepping.start()
    # sim.output.phys_fields.plot()

    fld.show()



"""NS2D equations over a sphere (:mod:`fluidsim.solvers.sphere.ns2d`)
=====================================================================

.. autoclass:: InfoSolverSphereNS2D
   :members:
   :private-members:

.. autoclass:: SimulSphereNS2D
   :members:
   :private-members:

"""

from fluidsim.base.sphericalharmo.solver import (
    InfoSolverSphericalHarmo,
    SimulSphericalHarmo,
)

from fluidsim.solvers.ns2d.solver import compute_Frot, SetOfVariables


class InfoSolverSphereNS2D(InfoSolverSphericalHarmo):
    """Contain the information on a base pseudo-spectral solver."""

    def _init_root(self):
        """Init. `self` by writting the information on the solver."""

        super()._init_root()

        here = "fluidsim.solvers.sphere.ns2d"

        self.module_name = here + ".solver"
        self.class_name = "SimulSphereNS2D"
        self.short_name = "sphere.ns2d"


# self.classes.State.module_name = here + '.state'
# self.classes.State.class_name = 'StateSphericalHarmo'

# self.classes.Output.module_name = here + '.output'
# self.classes.Output.class_name = 'Output'


class SimulSphereNS2D(SimulSphericalHarmo):
    """Pseudo-spectral base solver."""

    InfoSolver = InfoSolverSphereNS2D

    # @staticmethod
    # def _complete_params_with_default(params):
    #     """Complete the `params` container (static method)."""
    #     SimulSphericalHarmo._complete_params_with_default(params)
    #     # Coriolis parameter (not implemented)
    #     attribs = {'f': 0.}
    #     params._set_attribs(attribs)

    def tendencies_nonlin(self, state_spect=None, old=None):
        r"""Compute the nonlinear tendencies.

        Parameters
        ----------

        state_spect : :class:`fluidsim.base.setofvariables.SetOfVariables`
            optional

            Array containing the state, i.e. the vorticity, in Fourier
            space.  If `state_spect`, the variables vorticity and the
            velocity are computed from it, otherwise, they are taken
            from the global state of the simulation, `self.state`.

            These two possibilities are used during the Runge-Kutta
            time-stepping.

        Returns
        -------

        tendencies_sh : :class:`fluidsim.base.setofvariables.SetOfVariables`
            An array containing the tendencies for the vorticity.

        Notes
        -----

        .. |p| mathmacro:: \partial

        The 2D Navier-Stokes equation can be written

        .. math:: \p_t \hat\zeta = \hat N(\zeta) - \sigma(k) \hat \zeta,

        This function compute the nonlinear term ("tendencies")
        :math:`N(\zeta) = - \mathbf{u}\cdot \mathbf{\nabla} \zeta`.

        """
        # the operator and the fast Fourier transform
        oper = self.oper

        # get or compute rot_sh, ux and uy
        if state_spect is None:
            rot_sh = self.state.state_spect.get_var("rot_sh")
            ux = self.state.state_phys.get_var("ux")
            uy = self.state.state_phys.get_var("uy")
        else:
            rot_sh = state_spect.get_var("rot_sh")
            ux, uy = oper.vec_from_rotsh(rot_sh)

        # "px" like $\partial_x$
        px_rot, py_rot = oper.gradf_from_fsh(rot_sh)

        Frot = compute_Frot(ux, uy, px_rot, py_rot, beta=0.0)

        if old is None:
            tendencies_sh = SetOfVariables(like=self.state.state_spect)
        else:
            tendencies_sh = old
        Frot_sh = tendencies_sh.get_var("rot_sh")
        oper.sht_as_arg(Frot, Frot_sh)

        # oper.dealiasing(Frot_sh)

        T_rot = (Frot_sh.conj() * rot_sh).real
        print(
            ("sum(T_rot) = {:9.4e} ; sum(abs(T_rot)) = {:9.4e}").format(
                self.oper.sum_wavenumbers(T_rot),
                self.oper.sum_wavenumbers(abs(T_rot)),
            )
        )

        if self.params.forcing.enable:
            tendencies_sh += self.forcing.get_forcing()

        return tendencies_sh


Simul = SimulSphereNS2D

if __name__ == "__main__":

    params = Simul.create_default_params()

    params.short_name_type_run = "test"

    params.init_fields.type = "noise"

    params.time_stepping.USE_CFL = True
    params.time_stepping.t_end = 10.0
    # params.time_stepping.deltat0 = 0.1

    params.output.periods_save.phys_fields = 0.25

    sim = Simul(params)
    sim.time_stepping.start()



"""Spatial means (:mod:`fluidsim.solvers.ns2d.strat.output.spatial_means`)
==========================================================================

.. autoclass:: SpatialMeansNS2DStrat
   :members:
   :private-members:

"""

import os
import numpy as np
import matplotlib.pyplot as plt

from math import pi
from fluiddyn.util import mpi

from fluidsim.base.output.spatial_means import SpatialMeansBase


class SpatialMeansNS2DStrat(SpatialMeansBase):
    """Spatial means output stratified fluid"""

    def _save_one_time(self):
        tsim = self.sim.time_stepping.t
        self.t_last_save = tsim

        # Compute the kinetic, potential energy and enstrophy
        energyK_fft, energyA_fft = self.output.compute_energies_fft()
        energy_fft = self.output.compute_energy_fft()
        energy = self.sum_wavenumbers(energy_fft)
        energyK = self.sum_wavenumbers(energyK_fft)
        energyA = self.sum_wavenumbers(energyA_fft)

        enstrophy_fft = self.output.compute_enstrophy_fft()
        enstrophy = self.sum_wavenumbers(enstrophy_fft)

        # Compute energy shear modes
        COND_SHEAR_MODES = abs(self.sim.oper.KX) == 0.0
        energy_shear_modes = np.sum(energy_fft[COND_SHEAR_MODES])
        if mpi.nb_proc > 1:
            energy_shear_modes = mpi.comm.reduce(
                energy_shear_modes, op=mpi.MPI.SUM, root=0
            )

        # Dissipation rate kinetic and potential energy (kappa = viscosity)
        f_d, f_d_hypo = self.sim.compute_freq_diss()

        epsK = self.sum_wavenumbers(f_d * 2.0 * energyK_fft)
        epsK_hypo = self.sum_wavenumbers(f_d_hypo * 2.0 * energyK_fft)

        epsA = self.sum_wavenumbers(f_d * 2.0 * energyA_fft)
        epsA_hypo = self.sum_wavenumbers(f_d_hypo * 2.0 * energyA_fft)

        epsZ = self.sum_wavenumbers(f_d * 2.0 * enstrophy_fft)
        epsZ_hypo = self.sum_wavenumbers(f_d_hypo * 2.0 * enstrophy_fft)

        # Injection energy if forcing is True
        if self.sim.params.forcing.enable:
            deltat = self.sim.time_stepping.deltat
            Frot_fft = self.sim.forcing.get_forcing().get_var("rot_fft")
            Fb_fft = self.sim.forcing.get_forcing().get_var("b_fft")
            Fx_fft, Fy_fft = self.vecfft_from_rotfft(Frot_fft)

            rot_fft = self.sim.state.state_spect.get_var("rot_fft")
            b_fft = self.sim.state.state_spect.get_var("b_fft")

            ux_fft, uy_fft = self.vecfft_from_rotfft(rot_fft)

            PZ1_fft = np.real(rot_fft.conj() * Frot_fft)
            PZ2_fft = abs(Frot_fft) ** 2

            PZ1 = self.sum_wavenumbers(PZ1_fft)
            PZ2 = deltat / 2 * self.sum_wavenumbers(PZ2_fft)

            PK1_fft = np.real(ux_fft.conj() * Fx_fft + uy_fft.conj() * Fy_fft)
            PK2_fft = abs(Fx_fft) ** 2 + abs(Fy_fft) ** 2

            PK1 = self.sum_wavenumbers(PK1_fft)
            PK2 = deltat / 2 * self.sum_wavenumbers(PK2_fft)

            PA1_fft = np.real(b_fft.conj() * Fb_fft)
            PA2_fft = abs(Fb_fft) ** 2

            PA1 = self.sum_wavenumbers(PA1_fft) / self.params.N ** 2
            PA2 = deltat / 2 / self.params.N ** 2 * self.sum_wavenumbers(PA2_fft)

        if mpi.rank == 0:
            epsK_tot = epsK + epsK_hypo

            self.file.write(f"####\ntime = {tsim:11.5e}\n")
            to_print = (
                "Z    = {:11.5e} \n"
                "E    = {:11.5e} ; EK         = {:11.5e} ; EA        = {:11.5e} \n"
                "epsA = {:11.5e} ; epsA_hypo  = {:11.5e} ; epsA_tot  = {:11.5e} \n"
                "epsK = {:11.5e} ; epsK_hypo  = {:11.5e} ; epsK_tot  = {:11.5e} \n"
                "epsZ = {:11.5e} ; epsZ_hypo = {:11.5e} ; epsZ_tot = {:11.5e} \n"
                "E_shear = {:11.5e} \n"
            ).format(
                enstrophy,
                energy,
                energyK,
                energyA,
                epsA,
                epsA_hypo,
                epsA + epsA_hypo,
                epsK,
                epsK_hypo,
                epsK + epsK_hypo,
                epsZ,
                epsZ_hypo,
                epsZ + epsZ_hypo,
                energy_shear_modes,
            )
            self.file.write(to_print)

            if self.sim.params.forcing.enable:
                PK_tot = PK1 + PK2
                to_print = (
                    "PK1  = {:11.5e} ; PK2       = {:11.5e} ; PK_tot   = {:11.5e} \n"
                    "PA1  = {:11.5e} ; PA2       = {:11.5e} ; PA_tot   = {:11.5e} \n"
                    "PZ1  = {:11.5e} ; PZ2       = {:11.5e} ; PZ_tot   = {:11.5e} \n"
                ).format(
                    PK1, PK2, PK1 + PK2, PA1, PA2, PA1 + PA2, PZ1, PZ2, PZ1 + PZ2
                )
                self.file.write(to_print)

            self.file.flush()
            os.fsync(self.file.fileno())

        if self.has_to_plot and mpi.rank == 0:

            self.axe_a.plot(tsim, energy, "k.")

            self.axe_b.plot(tsim, epsK_tot, "k.")
            if self.sim.params.forcing.enable:
                self.axe_b.plot(tsim, PK_tot, "m.")

            if tsim - self.t_last_show >= self.period_show:
                self.t_last_show = tsim
                fig = self.axe_a.get_figure()
                fig.canvas.draw()

    def load(self):
        """Generates a dictionary with the output values"""
        dict_results = {"name_solver": self.output.name_solver}

        with open(self.path_file) as file_means:
            lines = file_means.readlines()

        lines_t = []
        lines_Z = []
        lines_E = []
        if self.sim.params.forcing.enable:
            lines_PK = []
            lines_PA = []
            lines_PZ = []
        lines_epsK = []
        lines_epsZ = []
        lines_epsA = []
        lines_E_shear = []

        for il, line in enumerate(lines):
            if line.startswith("time ="):
                lines_t.append(line)
            if line.startswith("Z    ="):
                lines_Z.append(line)
            if line.startswith("E    ="):
                lines_E.append(line)

            if self.sim.params.forcing.enable:
                if line.startswith("PK1  ="):
                    lines_PK.append(line)
                if line.startswith("PA1  ="):
                    lines_PA.append(line)
                if line.startswith("PZ1  ="):
                    lines_PZ.append(line)
            if line.startswith("epsK ="):
                lines_epsK.append(line)
            if line.startswith("epsZ ="):
                lines_epsZ.append(line)
            if line.startswith("epsA ="):
                lines_epsA.append(line)
            if line.startswith("E_shear ="):
                lines_E_shear.append(line)
        nt = len(lines_t)

        t = np.empty(nt)
        Z = np.empty(nt)
        E = np.empty(nt)
        EK = np.empty(nt)
        EA = np.empty(nt)

        if self.sim.params.forcing.enable:
            PK1 = np.empty(nt)
            PK2 = np.empty(nt)
            PK_tot = np.empty(nt)
            PA1 = np.empty(nt)
            PA2 = np.empty(nt)
            PA_tot = np.empty(nt)
            PZ1 = np.empty(nt)
            PZ2 = np.empty(nt)
            PZ_tot = np.empty(nt)

        epsK = np.empty(nt)
        epsK_hypo = np.empty(nt)
        epsK_tot = np.empty(nt)
        epsZ = np.empty(nt)
        epsZ_hypo = np.empty(nt)
        epsZ_tot = np.empty(nt)
        epsA = np.empty(nt)
        epsA_hypo = np.empty(nt)
        epsA_tot = np.empty(nt)
        E_shear = np.empty(nt)

        for il in range(nt):
            line = lines_t[il]
            words = line.split()
            t[il] = float(words[2])

            line = lines_Z[il]
            words = line.split()
            Z[il] = float(words[2])

            line = lines_E[il]
            words = line.split()
            E[il] = float(words[2])
            EK[il] = float(words[6])
            EA[il] = float(words[10])

            if self.sim.params.forcing.enable:
                line = lines_PK[il]
                words = line.split()
                PK1[il] = float(words[2])
                PK2[il] = float(words[6])
                PK_tot[il] = float(words[10])

                line = lines_PA[il]
                words = line.split()
                PA1[il] = float(words[2])
                PA2[il] = float(words[6])
                PA_tot[il] = float(words[10])

                line = lines_PZ[il]
                words = line.split()
                PZ1[il] = float(words[2])
                PZ2[il] = float(words[6])
                PZ_tot[il] = float(words[10])

            line = lines_epsK[il]
            words = line.split()
            epsK[il] = float(words[2])
            epsK_hypo[il] = float(words[6])
            epsK_tot[il] = float(words[10])

            line = lines_epsZ[il]
            words = line.split()
            epsZ[il] = float(words[2])
            epsZ_hypo[il] = float(words[6])
            epsZ_tot[il] = float(words[10])

            line = lines_epsA[il]
            words = line.split()
            epsA[il] = float(words[2])
            epsA_hypo[il] = float(words[6])
            epsA_tot[il] = float(words[10])

            line = lines_E_shear[il]
            words = line.split()
            E_shear[il] = float(words[2])

        dict_results["t"] = t
        dict_results["Z"] = Z

        dict_results["E"] = E
        dict_results["EK"] = EK
        dict_results["EA"] = EA

        if self.sim.params.forcing.enable:
            dict_results["PK1"] = PK1
            dict_results["PK2"] = PK2
            dict_results["PK_tot"] = PK_tot

            dict_results["PA1"] = PA1
            dict_results["PA2"] = PA2
            dict_results["PA_tot"] = PA_tot

            dict_results["PZ1"] = PZ1
            dict_results["PZ2"] = PZ2
            dict_results["PZ_tot"] = PZ_tot

        dict_results["epsK"] = epsK
        dict_results["epsK_hypo"] = epsK_hypo
        dict_results["epsK_tot"] = epsK_tot

        dict_results["epsZ"] = epsZ
        dict_results["epsZ_hypo"] = epsZ_hypo
        dict_results["epsZ_tot"] = epsZ_tot

        dict_results["epsA"] = epsA
        dict_results["epsA_hypo"] = epsA_hypo
        dict_results["epsA_tot"] = epsA_tot
        dict_results["E_shear"] = E_shear

        return dict_results

    def plot_energy_shear_modes(self):
        """
        Plots energy shear modes and total energy.

        """
        if mpi.rank != 0:
            return

        dict_results = self.load()

        times = dict_results["t"]
        E = dict_results["E"]
        E_shear_modes = dict_results["E_shear"]

        fig, ax = plt.subplots()
        ax.set_xlabel("Times")
        ax.set_ylabel("Energy")

        ax.plot(times, E, label=r"$E_{total}$")
        ax.plot(times, E_shear_modes, label=r"$E_{shear}$")
        ax.plot(times, E - E_shear_modes, label=r"$E_{total} - E_{shear}$")

        ax.legend()

        fig.tight_layout()

    def plot_dt_energy(self):
        """
        Checks if dE/dt = energy_injection - energy_dissipation.

        """
        if mpi.rank != 0:
            return

        dict_results = self.load()

        times = dict_results["t"]
        E = dict_results["E"]

        epsK_tot = dict_results["epsK_tot"]
        epsA_tot = dict_results["epsA_tot"]
        eps_tot = epsK_tot + epsA_tot

        dtE = np.diff(E) / np.diff(times)
        times_dtE = (times[1:] + times[:-1]) / 2

        if self.sim.params.forcing.enable:
            PK = dict_results["PK_tot"]
            PA = dict_results["PA_tot"]
            P = PK + PA
            model = P - eps_tot
        else:
            model = -eps_tot

        fig, axes = plt.subplots(2)
        ax = axes[0]
        ax.plot(times_dtE, dtE, label="$dE/dt$")

        times[:-1] += np.diff(times) / 2
        ax.plot(times[1:], model[1:], label=r"$P - \epsilon$")
        ax.legend()

        ax = axes[1]
        if self.sim.params.forcing.enable:
            ax.plot(times, PA + PK, "k", label="P")
            ax.plot(times, PA, "b", label="PA")
            ax.plot(times, PK, "r", label="PK")
            ax.plot(times, eps_tot, "k:", label=r"$\epsilon$")
        ax.set_xlabel("t")
        ax.legend()

        fig.tight_layout()

    def plot_energy(self):
        """Plots the energy."""
        if mpi.rank != 0:
            return

        pforcing = self.params.forcing
        dict_results = self.load()

        t = dict_results["t"]
        E = dict_results["E"]
        EK = dict_results["EK"]
        EA = dict_results["EA"]

        # Computation forcing wave-number k_f
        nkmax = pforcing.nkmax_forcing
        nkmin = pforcing.nkmin_forcing
        k_f = ((nkmax + nkmin) / 2) * max(
            2 * pi / self.sim.params.oper.Lx, 2 * pi / self.sim.params.oper.Ly
        )

        # Normalization by E_f
        if pforcing.key_forced == "ap_fft":
            E_f = pforcing.forcing_rate ** (2.0 / 7) * (2 * pi / k_f) ** (
                10.0 / 7
            )
        else:
            E_f = pforcing.forcing_rate ** (2.0 / 3) * (2 * pi / k_f) ** 2

        fig, ax = plt.subplots()
        ax.set_xlabel(r"$t/t_f$")
        ax.set_ylabel(r"$E/E_f$")
        ax.plot(t[1:], E[1:] / E_f, label="E", color="b")
        ax.plot(t[1:], EK[1:] / E_f, label="EK", color="r")
        ax.plot(t[1:], EA[1:] / E_f, label="EA", color="g")

        ax.legend()

        fig.tight_layout()

    def plot(self):
        if mpi.rank != 0:
            return

        dict_results = self.load()

        t = dict_results["t"]
        E = dict_results["E"]
        Z = dict_results["Z"]

        epsK = dict_results["epsK"]
        epsK_hypo = dict_results["epsK_hypo"]
        epsK_tot = dict_results["epsK_tot"]

        epsZ = dict_results["epsZ"]
        epsZ_hypo = dict_results["epsZ_hypo"]
        epsZ_tot = dict_results["epsZ_tot"]

        epsA = dict_results["epsA"]
        epsA_hypo = dict_results["epsA_hypo"]
        epsA_tot = dict_results["epsA_tot"]

        fig, (ax1, ax2) = plt.subplots(2)

        ax1.set_title("Energy and enstrophy")
        ax1.set_ylabel("$E(t)$")
        ax1.plot(t, E, "k")

        ax2.set_ylabel("$Z(t)$")
        ax2.set_xlabel("$t$")
        ax2.plot(t, Z, "k")
        fig.tight_layout()

        fig, (ax1, ax2) = plt.subplots(2)
        ax1.set_title("Dissipation of energy and enstrophy")
        ax1.set_ylabel(r"$\epsilon (t) = \epsilon_K + \epsilon_A$")

        if self.sim.params.nu_m4:
            ax1.plot(t, epsK + epsA, "r", label=r"$\epsilon$")

        ax1.plot(t, epsK_tot + epsA_tot, "k", label=r"$\epsilon_{tot}$")

        ax2.set_xlabel("$t$")
        ax2.set_ylabel(r"$\epsilon_Z(t)$")

        ax2.plot(t, epsZ_tot, "k", label=r"$\epsilon_Z$")

        # If true hypo-viscosity...
        if self.sim.params.nu_m4:
            ax2.plot(t, epsZ, "r")
            ax1.plot(
                t,
                epsK_hypo + epsA_hypo,
                color="g",
                label=r"$\epsilon_{hypo}$",
                linewidth=2,
            )
            ax2.plot(t, epsZ_hypo, "g")

        if self.sim.params.forcing.enable:
            PK = dict_results["PK_tot"]
            PA = dict_results["PA_tot"]
            P = PK + PA
            PZ_tot = dict_results["PZ_tot"]
            ax1.plot(t, P, "c", label="P")
            ax2.plot(t, PZ_tot, "c", label=r"$P_Z$")
            ax1.set_ylabel("P_E(t), epsK(t)")
            ax2.set_ylabel("P_Z(t), epsZ(t)")

        ax1.legend()
        ax2.legend()
        fig.tight_layout()



"""Standard output (:mod:`fluidsim.solvers.plate2d.output.print_stdout`)
==============================================================================


Provides:

.. autoclass:: PrintStdOutPlate2D
   :members:
   :private-members:

"""

from fluidsim.solvers.ns2d.output.print_stdout import PrintStdOutNS2D


class PrintStdOutPlate2D(PrintStdOutNS2D):
    """Used to print in both the stdout and the stdout.txt file, and also
    to print simple info on the current state of the simulation.

    """



"""Operators finite differences (:mod:`fluidsim.operators.op_finitediff2d`)
===========================================================================

Provides:

.. autoclass:: OperatorFiniteDiff2DPeriodic
   :members:
   :private-members:

"""

from .op_finitediff1d import OperatorFiniteDiff1DPeriodic


class OperatorFiniteDiff2DPeriodic(OperatorFiniteDiff1DPeriodic):
    @staticmethod
    def _complete_params_with_default(params):
        """This static method is used to complete the *params* container."""

        attribs = {"nx": 48, "ny": 48, "Lx": 8, "Ly": 8}
        params._set_child("oper", attribs=attribs)

    def __init__(self, params=None):

        Lx = float(params.oper.Lx)
        Ly = float(params.oper.Ly)

        if not params.ONLY_COARSE_OPER:
            nx = int(params.oper.nx)
            ny = int(params.oper.ny)
        else:
            nx = 4
            ny = 4

        self.nx = nx
        self.ny = ny
        self.shape = [ny, nx]
        size = nx * ny
        self.size = size
        self.lx = self.Lx = Lx
        self.ly = self.Ly = Ly
        self.deltax = Lx / nx
        self.deltay = Ly / ny
        dx = self.deltax
        dy = self.deltay

        self.xs = np.linspace(0, Lx, nx)
        self.ys = np.linspace(0, Ly, ny)

        def func_i1_mat(i0_mat, iv):
            i1 = i0_mat % nx
            i0 = i0_mat // nx
            if iv == 0:
                i1_mat = i0 * nx + (i1 + 1) % nx
            elif iv == 1:
                i1_mat = i0 * nx + (i1 - 1) % nx
            else:
                raise ValueError("Shouldn" "t be here...")

            return i1_mat

        values = np.array([1, -1]) / (2 * dx)
        self.sparse_px = self._create_sparse(values, func_i1_mat)

        def func_i1_mat(i0_mat, iv):
            i1 = i0_mat % nx
            i0 = i0_mat // nx
            if iv == 0:
                i1_mat = i0_mat
            elif iv == 1:
                i1_mat = i0 * nx + (i1 + 1) % nx
            elif iv == 2:
                i1_mat = i0 * nx + (i1 - 1) % nx
            else:
                raise ValueError("Shouldn" "t be here...")

            return i1_mat

        values = np.array([-2, 1, 1]) / dx ** 2
        self.sparse_pxx = self._create_sparse(values, func_i1_mat)

        def func_i1_mat(i0_mat, iv):
            i1 = i0_mat % nx
            i0 = i0_mat // nx
            if iv == 0:
                i1_mat = ((i0 + 1) * nx) % size + i1
            elif iv == 1:
                i1_mat = ((i0 - 1) * nx) % size + i1
            else:
                raise ValueError("Shouldn" "t be here...")

            return i1_mat

        values = np.array([1, -1]) / (2 * dy)
        self.sparse_py = self._create_sparse(values, func_i1_mat)

        def func_i1_mat(i0_mat, iv):
            i1 = i0_mat % nx
            i0 = i0_mat // nx
            if iv == 0:
                i1_mat = i0_mat
            elif iv == 1:
                i1_mat = ((i0 + 1) * nx) % size + i1
            elif iv == 2:
                i1_mat = ((i0 - 1) * nx) % size + i1
            else:
                raise ValueError("Shouldn" "t be here...")

            return i1_mat

        values = np.array([-2, 1, 1]) / dx ** 2
        self.sparse_pyy = self._create_sparse(values, func_i1_mat)

    def _create_sparse(self, values, func_i1_mat):
        size = self.size
        nb_values = len(values)
        data = np.empty(size * nb_values)
        i0s = np.empty(size * nb_values)
        i1s = np.empty(size * nb_values)

        for i0_mat in range(size):
            for iv, v in enumerate(values):
                data[nb_values * i0_mat + iv] = v
                i0s[nb_values * i0_mat + iv] = i0_mat
                i1s[nb_values * i0_mat + iv] = func_i1_mat(i0_mat, iv)
        return sparse.coo_matrix((data, (i0s, i1s)), shape=(size, size))

    def py(self, a):
        return self.sparse_py.dot(a.flat)

    def pyy(self, a):
        return self.sparse_pyy.dot(a.flat)

    def produce_str_describing_oper(self):
        """Produce a string describing the operator."""
        if (self.Lx / np.pi).is_integer():
            str_Lx = repr(int(self.Lx // np.pi)) + "pi"
        else:
            str_Lx = f"{self.Lx:.3f}".rstrip("0")
        if (self.Ly / np.pi).is_integer():
            str_Ly = repr(int(self.Ly // np.pi)) + "pi"
        else:
            str_Ly = f"{self.Ly:.3f}".rstrip("0")
        return ("L=" + str_Lx + "x" + str_Ly + "_{}x{}").format(self.nx, self.ny)

    def produce_long_str_describing_oper(self):
        """Produce a string describing the operator."""
        if (self.Lx / np.pi).is_integer():
            str_Lx = repr(int(self.Lx // np.pi)) + "pi"
        else:
            str_Lx = f"{self.Lx:.3f}".rstrip("0")
        if (self.Ly / np.pi).is_integer():
            str_Ly = repr(int(self.Ly // np.pi)) + "pi"
        else:
            str_Ly = f"{self.Ly:.3f}".rstrip("0")
        return (
            "Finite difference operator 2D,\n"
            "nx = {:6d} ; ny = {:6d}\n".format(self.nx, self.ny)
            + "Lx = "
            + str_Lx
            + " ; Ly = "
            + str_Ly
            + "\n"
        )


if __name__ == "__main__":
    nx = 3
    ny = 3
    oper = OperatorFiniteDiff2DPeriodic([ny, nx], [nx / 2.0, ny / 2.0])
    a = np.arange(nx * ny).reshape([ny, nx])



"""Spatial means (:mod:`fluidsim.solvers.ns3d.strat.output.spatial_means`)
==========================================================================

.. autoclass:: SpatialMeansNS3DStrat
   :members:
   :private-members:

"""

import os
import numpy as np


from fluiddyn.util import mpi

from fluidsim.solvers.ns3d.output.spatial_means import SpatialMeansNS3D


class SpatialMeansNS3DStrat(SpatialMeansNS3D):
    """Spatial means output."""

    def __init__(self, output):
        self.one_over_N2 = 1.0 / output.sim.params.N ** 2
        super().__init__(output)

    def _save_one_time(self):
        tsim = self.sim.time_stepping.t
        self.t_last_save = tsim
        nrj_A, nrj_Kz, nrj_Khr, nrj_Khd = self.output.compute_energies_fft()
        energyK_fft = nrj_Kz + nrj_Khr + nrj_Khd
        # shear modes
        COND_SHEAR = self.oper.Kx ** 2 + self.oper.Ky ** 2 == 0.0
        nrj_Khs = nrj_Khr * COND_SHEAR
        nrj_As = nrj_A * COND_SHEAR
        energyA_fft = nrj_A
        nrj_A = self.sum_wavenumbers(nrj_A)
        nrj_As = self.sum_wavenumbers(nrj_As)
        nrj_Kz = self.sum_wavenumbers(nrj_Kz)
        nrj_Khs = self.sum_wavenumbers(nrj_Khs)
        nrj_Khr = self.sum_wavenumbers(nrj_Khr)
        nrj_Khr = nrj_Khr - nrj_Khs
        nrj_Khd = self.sum_wavenumbers(nrj_Khd)
        energy = nrj_A + nrj_Kz + nrj_Khr + nrj_Khd + nrj_Khs

        f_d, f_d_hypo = self.sim.compute_freq_diss()
        epsK = self.sum_wavenumbers(f_d * 2 * energyK_fft)
        epsK_hypo = self.sum_wavenumbers(f_d_hypo * 2 * energyK_fft)
        epsA = self.sum_wavenumbers(f_d * 2 * energyA_fft)
        epsA_hypo = self.sum_wavenumbers(f_d_hypo * 2 * energyA_fft)

        if self.sim.params.nu_4 > 0.0:
            f_d4 = self.params.nu_4 * self.oper.K4
            epsK4 = self.sum_wavenumbers(f_d4 * 2 * energyK_fft)
            epsA4 = self.sum_wavenumbers(f_d4 * 2 * energyA_fft)
            del f_d4

        if self.sim.params.nu_8 > 0.0:
            f_d8 = self.params.nu_8 * self.oper.K8
            epsK8 = self.sum_wavenumbers(f_d8 * 2 * energyK_fft)
            epsA8 = self.sum_wavenumbers(f_d8 * 2 * energyA_fft)
            del f_d8

        if self.sim.params.forcing.enable:
            deltat = self.sim.time_stepping.deltat
            forcing_fft = self.sim.forcing.get_forcing()

            fx_fft = forcing_fft.get_var("vx_fft")
            fy_fft = forcing_fft.get_var("vy_fft")
            fz_fft = forcing_fft.get_var("vz_fft")
            fb_fft = forcing_fft.get_var("b_fft")

            get_var = self.sim.state.state_spect.get_var
            vx_fft = get_var("vx_fft")
            vy_fft = get_var("vy_fft")
            vz_fft = get_var("vz_fft")
            b_fft = get_var("b_fft")

            PK1_fft = np.real(
                vx_fft.conj() * fx_fft
                + vy_fft.conj() * fy_fft
                + vz_fft.conj() * fz_fft
            )
            PK2_fft = abs(fx_fft) ** 2 + abs(fy_fft) ** 2 + abs(fz_fft) ** 2

            PK1 = self.sum_wavenumbers(np.ascontiguousarray(PK1_fft))
            PK2 = self.sum_wavenumbers(PK2_fft) * deltat / 2

            PA1_fft = np.real(b_fft.conj() * fb_fft)
            PA2_fft = abs(fb_fft) ** 2

            PA1 = self.sum_wavenumbers(np.ascontiguousarray(PA1_fft))
            PA2 = self.sum_wavenumbers(PA2_fft) * deltat / 2

            PA1 *= self.one_over_N2
            PA2 *= self.one_over_N2

        if mpi.rank == 0:

            self.file.write(
                f"####\ntime = {tsim:11.5e}\n"
                f"E    = {energy:11.5e}\n"
                f"EA   = {nrj_A:11.5e} ; EKz   = {nrj_Kz:11.5e} ; "
                f"EKhr   = {nrj_Khr:11.5e} ; EKhd   = {nrj_Khd:11.5e} ; "
                f"EKhs   = {nrj_Khs:11.5e} ; EAs    = {nrj_As:11.5e}\n"
                f"epsK = {epsK:11.5e} ; epsK_hypo = {epsK_hypo:11.5e} ; "
                f"epsA = {epsA:11.5e} ; epsA_hypo = {epsA_hypo:11.5e} ; "
                f"eps_tot = {epsK + epsK_hypo + epsA + epsA_hypo:11.5e} \n"
            )

            if self.sim.params.nu_4 > 0.0:
                self.file.write(
                    f"epsK4 = {epsK4:11.5e} ; epsA4 = {epsA4:11.5e}\n"
                )

            if self.sim.params.nu_8 > 0.0:
                self.file.write(
                    f"epsK8 = {epsK8:11.5e} ; epsA8 = {epsA8:11.5e}\n"
                )

            if self.sim.params.forcing.enable:
                self.file.write(
                    f"PK1  = {PK1:11.5e} ; PK2       = {PK2:11.5e} ; "
                    f"PK_tot   = {PK1 + PK2:11.5e} \n"
                    f"PA1  = {PA1:11.5e} ; PA2       = {PA2:11.5e} ; "
                    f"PA_tot   = {PA1 + PA2:11.5e} \n"
                )

            self.file.flush()
            os.fsync(self.file.fileno())

        if self.has_to_plot and mpi.rank == 0:

            self.axe_a.plot(tsim, energy, "k.")

            # self.axe_b.plot(tsim, epsK_tot, 'k.')
            # if self.sim.params.forcing.enable:
            #     self.axe_b.plot(tsim, PK_tot, 'm.')

            if tsim - self.t_last_show >= self.period_show:
                self.t_last_show = tsim
                fig = self.axe_a.get_figure()
                fig.canvas.draw()

    def load(self):
        results = {"name_solver": self.output.name_solver}

        with open(self.path_file) as file_means:
            lines = file_means.readlines()

        lines_t = []
        lines_E = []
        lines_EA = []
        lines_PK = []
        lines_PA = []
        lines_epsK = []
        lines_epsK4 = []
        lines_epsK8 = []

        for il, line in enumerate(lines):
            if line.startswith("time ="):
                lines_t.append(line)
            elif line.startswith("E    ="):
                lines_E.append(line)
            elif line.startswith("EA   ="):
                lines_EA.append(line)
            elif line.startswith("PK1  ="):
                lines_PK.append(line)
            elif line.startswith("PA1  ="):
                lines_PA.append(line)
            elif line.startswith("epsK ="):
                lines_epsK.append(line)
            elif line.startswith("epsK4 ="):
                lines_epsK4.append(line)
            elif line.startswith("epsK8 ="):
                lines_epsK8.append(line)

        nt = len(lines_t)
        if nt > 1:
            nt -= 1

        # support files saved without EAs
        words = lines_EA[0].split()
        try:
            words[22]
        except IndexError:
            EAs_saved = False
        else:
            EAs_saved = True

        t = np.empty(nt)
        E = np.empty(nt)
        EA = np.empty(nt)
        EKz = np.empty(nt)
        EKhr = np.empty(nt)
        EKhd = np.empty(nt)
        EKhs = np.empty(nt)
        if EAs_saved:
            EAs = np.empty(nt)
        PK1 = np.zeros(nt)
        PK2 = np.zeros(nt)
        PK_tot = np.zeros(nt)
        PA1 = np.zeros(nt)
        PA2 = np.zeros(nt)
        PA_tot = np.zeros(nt)
        epsK = np.empty(nt)
        epsK_hypo = np.empty(nt)
        epsA = np.empty(nt)
        epsA_hypo = np.empty(nt)
        eps_tot = np.empty(nt)

        if lines_epsK4:
            epsK4 = np.empty(nt)
            epsA4 = np.empty(nt)

        if lines_epsK8:
            epsK8 = np.empty(nt)
            epsA8 = np.empty(nt)

        for il in range(nt):
            line = lines_t[il]
            words = line.split()
            t[il] = float(words[2])

            line = lines_E[il]
            words = line.split()
            E[il] = float(words[2])

            line = lines_EA[il]
            words = line.split()
            EA[il] = float(words[2])
            EKz[il] = float(words[6])
            EKhr[il] = float(words[10])
            EKhd[il] = float(words[14])
            EKhs[il] = float(words[18])
            if EAs_saved:
                EAs[il] = float(words[22])

            if self.sim.params.forcing.enable:
                line = lines_PK[il]
                words = line.split()
                PK1[il] = float(words[2])
                PK2[il] = float(words[6])
                PK_tot[il] = float(words[10])

                line = lines_PA[il]
                words = line.split()
                PA1[il] = float(words[2])
                PA2[il] = float(words[6])
                PA_tot[il] = float(words[10])

            line = lines_epsK[il]
            words = line.split()
            epsK[il] = float(words[2])
            epsK_hypo[il] = float(words[6])
            epsA[il] = float(words[10])
            epsA_hypo[il] = float(words[14])
            eps_tot[il] = float(words[18])

            if lines_epsK4:
                line = lines_epsK4[il]
                words = line.split()
                epsK4[il] = float(words[2])
                epsA4[il] = float(words[6])

            if lines_epsK8:
                line = lines_epsK8[il]
                words = line.split()
                epsK8[il] = float(words[2])
                epsA8[il] = float(words[6])

        results["t"] = t
        results["E"] = E
        results["EA"] = EA
        results["EKz"] = EKz
        results["EKhr"] = EKhr
        results["EKhd"] = EKhd
        results["EKhs"] = EKhs
        if EAs_saved:
            results["EAs"] = EAs

        results["PK1"] = PK1
        results["PK2"] = PK2
        results["PK_tot"] = PK_tot

        results["PA1"] = PA1
        results["PA2"] = PA2
        results["PA_tot"] = PA_tot

        results["epsK"] = epsK
        results["epsK_hypo"] = epsK_hypo
        results["epsA"] = epsA
        results["epsA_hypo"] = epsA_hypo
        results["eps_tot"] = eps_tot

        if lines_epsK4:
            results["epsK4"] = epsK4
            results["epsA4"] = epsA4

        if lines_epsK8:
            results["epsK8"] = epsK8
            results["epsA8"] = epsA8

        return results

    def plot(self, plot_injection=True, plot_hyper=False):
        results = self.load()

        t = results["t"]
        E = results["E"]
        EA = results["EA"]
        EKz = results["EKz"]
        EKhr = results["EKhr"]
        EKhd = results["EKhd"]
        EKhs = results["EKhs"]
        EK = EKz + EKhr + EKhd + EKhs

        epsK = results["epsK"]
        epsK_hypo = results["epsK_hypo"]
        epsA = results["epsA"]
        epsA_hypo = results["epsA_hypo"]
        eps_tot = results["eps_tot"]

        # fig 1 : energies
        fig, ax = self.output.figure_axe()
        ax.set_title("Energy\n" + self.output.summary_simul)
        ax.set_ylabel("$E(t)$")
        ax.plot(t, E, "k", linewidth=2, label="$E$")
        ax.plot(t, EA, "b", label="$E_A$")
        ax.plot(t, EK, "r", label="$E_K$")
        ax.plot(t, EKhr, "r:", label="$E_{Khr}$")
        ax.plot(t, EKhs, "m:", label="$E_{Khs}$")

        try:
            EAs = results["EAs"]
        except KeyError:
            pass
        else:
            ax.plot(t, EAs, "g:", label="$E_{As}$")

        ax.legend()

        # figure 2 : dissipations
        fig, ax = self.output.figure_axe()
        ax.set_title("Dissipation of energy\n" + self.output.summary_simul)
        ax.set_ylabel(r"$\epsilon_K(t)$")

        def _plot(x, y, fmt, label=None, linewidth=1, zorder=10):
            ax.plot(x, y, fmt, label=label, linewidth=linewidth, zorder=zorder)

        _plot(t, epsK, "r", r"$\epsilon_K$")
        _plot(t, epsA, "b", r"$\epsilon_A$")
        _plot(t, eps_tot, "k", r"$\epsilon$", linewidth=2)

        eps_hypo = epsK_hypo + epsA_hypo
        if max(eps_hypo) > 0:
            _plot(t, eps_hypo, "g", r"$\epsilon_{hypo}$")

        if "epsK4" in results and plot_hyper:
            epsK4 = results["epsK4"]
            epsA4 = results["epsA4"]
            if not np.allclose(epsK, epsK4):
                _plot(t, epsK4, "r:", r"$\epsilon_{K4}$")
                _plot(t, epsA4, "b:", r"$\epsilon_{A4}$")

        if "epsK8" in results and plot_hyper:
            epsK8 = results["epsK8"]
            epsA8 = results["epsA8"]
            if not np.allclose(epsK, epsK8):
                _plot(t, epsK8, "r:", r"$\epsilon_{K8}$")
                _plot(t, epsA8, "b:", r"$\epsilon_{A8}$")

        if "PK_tot" in results and plot_injection:
            PK_tot = results["PK_tot"]
            PA_tot = results["PA_tot"]
            ax.plot(t, PK_tot, "r--", label=r"$P_K$", zorder=0)
            ax.plot(t, PA_tot, "b--", label=r"$P_A$", zorder=0)

        ax.legend()

    def get_dimless_numbers_versus_time(self):
        data = self.load()
        results = {"t": data["t"]}
        EKhr = data["EKhr"]
        EKhd = data["EKhd"]
        EKhs = data["EKhs"]
        epsK = data["epsK"]
        epsA = data["epsA"]

        epsK_hyper = np.zeros_like(epsK)

        Uh2 = EKhr + EKhd + EKhs

        N = self.params.N
        nu_2 = self.params.nu_2
        nu_4 = self.params.nu_4
        nu_8 = self.params.nu_8

        results["Fh"] = epsK / (Uh2 * N)

        if nu_2:
            results["R2"] = epsK / (nu_2 * N ** 2)
        if nu_4:
            results["R4"] = epsK * Uh2 / (nu_4 * N ** 4)
            epsK_hyper += data["epsK4"]
        if nu_8:
            results["R8"] = epsK * Uh2 ** 3 / (nu_8 * N ** 8)
            epsK_hyper += data["epsK8"]

        if nu_2 and (nu_4 or nu_8):
            epsK2 = epsK - epsK_hyper
            results["epsK2/epsK"] = epsK2 / epsK

        results["Gamma"] = epsA / epsK
        results["dimensional"] = {"Uh2": Uh2, "epsK": epsK}
        return results

    def get_dimless_numbers_averaged(self, tmin=0, tmax=None):
        numbers_vs_time = self.get_dimless_numbers_versus_time()
        times = numbers_vs_time["t"]
        itmin, itmax = _compute_indices_tmin_tmax(times, tmin, tmax)
        stop = itmax + 1

        result = {
            k: q[itmin:stop].mean()
            for k, q in numbers_vs_time.items()
            if k not in ["t", "dimensional"]
        }
        result["dimensional"] = {
            k: q[itmin:stop].mean()
            for k, q in numbers_vs_time["dimensional"].items()
        }
        return result

    def plot_dimless_numbers_versus_time(self, tmin=0, tmax=None):
        numbers_vs_time = self.get_dimless_numbers_versus_time()
        times = numbers_vs_time["t"]
        itmin, itmax = _compute_indices_tmin_tmax(times, tmin, tmax)
        stop = itmax + 1
        times = times[itmin:stop]

        fig, ax = self.output.figure_axe()

        for key, quantity in numbers_vs_time.items():
            if key in ["t", "dimensional"]:
                continue
            ax.plot(times, quantity[itmin:stop], label=key)

        ax.set_yscale("log")
        ax.set_xlabel("$t$")
        ax.set_title(f"dimensionless numbers\n{self.output.summary_simul}")

        fig.legend()


def _compute_indices_tmin_tmax(times, tmin, tmax):
    if tmax is None:
        itmax = len(times) - 1
    itmin = abs(times - tmin).argmin()
    return itmin, itmax



"""State for the NS2D solver (:mod:`fluidsim.solvers.ns2d.state`)
=================================================================

.. autoclass:: StateNS2D
   :members:
   :private-members:

"""

import numpy as np

from fluidsim.base.state import StatePseudoSpectral

from fluiddyn.util import mpi


class StateNS2D(StatePseudoSpectral):
    """State for the solver ns2d.

    Contains the variables corresponding to the state and handles the
    access to other fields.

    .. inheritance-diagram:: StateNS2D

    """

    @staticmethod
    def _complete_info_solver(info_solver):
        """Complete the `info_solver` container (static method)."""
        info_solver.classes.State._set_attribs(
            {
                "keys_state_spect": ["rot_fft"],
                "keys_state_phys": ["ux", "uy", "rot"],
                "keys_computable": [],
                "keys_phys_needed": ["rot"],
                "keys_linear_eigenmodes": ["rot_fft"],
            }
        )

    def __init__(self, sim, oper=None):

        super().__init__(sim, oper)

        self.field_tmp0 = np.empty_like(self.state_phys[0])
        self.field_tmp1 = np.empty_like(self.state_phys[0])
        self.field_tmp2 = np.empty_like(self.state_phys[0])
        self.field_tmp3 = np.empty_like(self.state_phys[0])

    def compute(self, key, SAVE_IN_DICT=True, RAISE_ERROR=True):
        """Compute and return a variable"""
        it = self.sim.time_stepping.it
        if key in self.vars_computed and it == self.it_computed[key]:
            return self.vars_computed[key]

        if key == "ux_fft":
            # not efficient!
            result = self.oper.fft2(self.state_phys.get_var("ux"))
        elif key == "uy_fft":
            # not efficient!
            result = self.oper.fft2(self.state_phys.get_var("uy"))
        elif key == "rot_fft":
            ux_fft = self.compute("ux_fft")
            uy_fft = self.compute("uy_fft")
            result = self.oper.rotfft_from_vecfft(ux_fft, uy_fft)
        elif key == "div_fft":
            ux_fft = self.compute("ux_fft")
            uy_fft = self.compute("uy_fft")
            result = self.oper.divfft_from_vecfft(ux_fft, uy_fft)
        elif key == "div":
            div_fft = self.compute("div_fft")
            result = self.oper.ifft2(div_fft)
        elif key == "q":
            rot = self.get_var("rot")
            result = rot
        else:
            to_print = 'Do not know how to compute "' + key + '".'
            if RAISE_ERROR:
                raise ValueError(to_print)

            else:
                if mpi.rank == 0:
                    print(to_print + "\nreturn an array of zeros.")

                result = self.oper.create_arrayX(value=0.0)

        if SAVE_IN_DICT:
            self.vars_computed[key] = result
            self.it_computed[key] = it

        return result

    def statephys_from_statespect(self):
        """Compute `state_phys` from `statespect`."""
        rot_fft = self.state_spect.get_var("rot_fft")
        ux_fft, uy_fft = self.oper.vecfft_from_rotfft(rot_fft)

        rot = self.state_phys.get_var("rot")
        ux = self.state_phys.get_var("ux")
        uy = self.state_phys.get_var("uy")

        self.oper.ifft_as_arg(rot_fft, rot)
        self.oper.ifft_as_arg(ux_fft, ux)
        self.oper.ifft_as_arg(uy_fft, uy)

    def statespect_from_statephys(self):
        """Compute `state_spect` from `state_phys`."""

        rot = self.state_phys.get_var("rot")
        rot_fft = self.state_spect.get_var("rot_fft")
        self.oper.fft_as_arg(rot, rot_fft)

    def init_from_rotfft(self, rot_fft):
        """Initialize the state from the variable `rot_fft`."""
        self.oper.dealiasing(rot_fft)
        self.state_spect.set_var("rot_fft", rot_fft)
        self.statephys_from_statespect()

    def init_from_uxfft(self, ux_fft):
        """Initialize the state from the variable `ux_fft`"""
        uy_fft = self.oper.create_arrayK(value=0.0)
        rot_fft = self.oper.rotfft_from_vecfft(ux_fft, uy_fft)
        self.init_from_rotfft(rot_fft)

    def init_from_uyfft(self, uy_fft):
        """Initialize the state from the variable `uy_fft`"""
        ux_fft = self.oper.create_arrayK(value=0.0)
        rot_fft = self.oper.rotfft_from_vecfft(ux_fft, uy_fft)
        self.init_from_rotfft(rot_fft)

    def init_statespect_from(self, **kwargs):
        """Initializes *state_spect* using arrays provided as keyword
        arguments.

        """
        if len(kwargs) == 1:
            key, arr = kwargs.popitem()
            if key == "rot_fft":
                self.init_from_rotfft(arr)
            elif key == "ux_fft":
                self.init_from_uxfft(arr)
            elif key == "uy_fft":
                self.init_from_uyfft(arr)
            else:
                super().init_statespect_from(**kwargs)
        else:
            super().init_statespect_from(**kwargs)



"""Dedalus solver (:mod:`fluidsim.base.dedalus.solver`)
=========================================================

Provides:

.. autoclass:: InfoSolverDedalus
   :members:
   :private-members:

.. autoclass:: SimulDedalus
   :members:
   :private-members:

"""


from fluidsim.base.solvers.base import SimulBase
from fluidsim.base.solvers.info_base import InfoSolverBase

from dedalus import public as dedalus


class InfoSolverDedalus(InfoSolverBase):
    """Contain the information on a Dedalus solver."""

    def _init_root(self):

        super()._init_root()

        mod = "fluidsim.base.dedalus"

        self.module_name = mod + ".solver"
        self.class_name = "SimulDedalus"
        self.short_name = "dedalus"

        classes = self.classes

        classes.State.module_name = mod + ".state"
        classes.State.class_name = "StateDedalus"

        classes.TimeStepping.module_name = mod + ".time_stepping"
        classes.TimeStepping.class_name = "TimeSteppingDedalus"

        classes.Operators.module_name = mod + ".operators"
        classes.Operators.class_name = "OperatorsDedalus2D"

        classes.Output.module_name = mod + ".output"
        classes.Output.class_name = "OutputDedalus"


class SimulDedalus(SimulBase):
    """A solver for Dedalus."""

    dedalus = dedalus

    InfoSolver = InfoSolverDedalus

    @staticmethod
    def _complete_params_with_default(params):
        """Complete the `params` container (static method)."""
        SimulBase._complete_params_with_default(params)
        attribs = {"prandtl": 1.0, "rayleigh": 1e6, "F": 1.0}
        params._set_attribs(attribs)

    def create_problem(self):
        params = self.params

        # 2D Boussinesq hydrodynamics
        problem = self.dedalus.IVP(
            self.oper.domain,
            variables=["p", "b", "vx", "vz", "dz_b", "dz_vx", "dz_vz"],
        )
        problem.meta["p", "b", "vx", "vz"]["z"]["dirichlet"] = True
        problem.parameters["P"] = (params.rayleigh * params.prandtl) ** (-1 / 2)
        problem.parameters["R"] = (params.rayleigh / params.prandtl) ** (-1 / 2)
        problem.parameters["F"] = params.F
        problem.add_equation("dx(vx) + dz_vz = 0")
        problem.add_equation(
            "dt(b) - P*(dx(dx(b)) + dz(dz_b)) - F*vz       = -(vx*dx(b) + vz*dz_b)"
        )
        problem.add_equation(
            "dt(vx) - R*(dx(dx(vx)) + dz(dz_vx)) + dx(p)     = -(vx*dx(vx) + vz*dz_vx)"
        )
        problem.add_equation(
            "dt(vz) - R*(dx(dx(vz)) + dz(dz_vz)) + dz(p) - b = -(vx*dx(vz) + vz*dz_vz)"
        )
        problem.add_equation("dz_b - dz(b) = 0")
        problem.add_equation("dz_vx - dz(vx) = 0")
        problem.add_equation("dz_vz - dz(vz) = 0")
        problem.add_bc("left(b) = 0")
        problem.add_bc("left(vx) = 0")
        problem.add_bc("left(vz) = 0")
        problem.add_bc("right(b) = 0")
        problem.add_bc("right(vx) = 0")
        problem.add_bc("right(vz) = 0", condition="(nx != 0)")
        problem.add_bc("right(p) = 0", condition="(nx == 0)")
        return problem

    def build_dedalus_solver(self):
        solver = self.problem.build_solver(self.dedalus.timesteppers.RK222)
        return solver

    def init_dedalus(self):
        self.problem = self.create_problem()

        if not self.params.ONLY_COARSE_OPER:
            self.dedalus_solver = self.build_dedalus_solver()


Simul = SimulDedalus

if __name__ == "__main__":

    params = Simul.create_default_params()

    params.short_name_type_run = "test"

    params.oper.nx = 48

    params.time_stepping.t_end = 10
    params.time_stepping.USE_CFL = True
    params.time_stepping.deltat0 = 2.0

    params.output.periods_print.print_stdout = 1e-15
    params.output.periods_save.phys_fields = 0.5

    sim = Simul(params)
    sim.time_stepping.start()



"""Physical fields output 3d (:mod:`fluidsim.base.output.phys_fields3d`)
========================================================================

Provides:

.. autoclass:: MoviesBasePhysFields3D
   :members:
   :private-members:

.. autoclass:: PhysFieldsBase3D
   :members:
   :private-members:

"""

import numpy as np
import matplotlib.pyplot as plt

from fluiddyn.util import mpi

from .phys_fields2d import MoviesBasePhysFields2D, PhysFieldsBase2D


def _get_xylabels_from_equation(equation):
    if equation.startswith("iz=") or equation.startswith("z="):
        xlabel = "x"
        ylabel = "y"
    elif equation.startswith("iy=") or equation.startswith("y="):
        xlabel = "x"
        ylabel = "z"
    elif equation.startswith("ix=") or equation.startswith("x="):
        xlabel = "y"
        ylabel = "z"
    else:
        raise NotImplementedError
    return xlabel, ylabel


class MoviesBasePhysFields3D(MoviesBasePhysFields2D):
    def _init_labels(self, xlabel=None, ylabel=None):
        """Initialize the labels."""
        if xlabel is None or ylabel is None:
            _xlabel, _ylabel = _get_xylabels_from_equation(self._equation)
        if xlabel is None:
            xlabel = _xlabel
        if ylabel is None:
            ylabel = _ylabel
        self.ax.set_xlabel(xlabel, fontdict=self.font)
        self.ax.set_ylabel(ylabel, fontdict=self.font)


class PhysFieldsBase3D(PhysFieldsBase2D):
    def __init__(self, output):
        super().__init__(output)
        self.set_equation_crosssection("iz=0")

    def _init_movies(self):
        self.movies = MoviesBasePhysFields3D(self.output, self)

    def set_equation_crosssection(self, equation):
        """Set the equation defining the cross-section.

        Parameters
        ----------

        equation : str

          The equation can be of the shape 'iz=2', 'z=1', ...

        """
        self._equation = equation
        self.movies._equation = equation

    def _get_grid1d(self, equation):

        if equation.startswith("iz=") or equation.startswith("z="):
            x_seq = self.oper.get_grid1d_seq("x")
            y_seq = self.oper.get_grid1d_seq("y")
        elif equation.startswith("iy=") or equation.startswith("y="):
            x_seq = self.oper.get_grid1d_seq("x")
            y_seq = self.oper.get_grid1d_seq("z")
        elif equation.startswith("ix=") or equation.startswith("x="):
            x_seq = self.oper.get_grid1d_seq("y")
            y_seq = self.oper.get_grid1d_seq("z")
        else:
            raise NotImplementedError

        return x_seq, y_seq

    def plot(
        self,
        field=None,
        time=None,
        QUIVER=True,
        vector="v",
        equation=None,
        nb_contours=20,
        type_plot="contourf",
        vmin=None,
        vmax=None,
        cmap=None,
        numfig=None,
        SCALED=True,
    ):
        """Plot a field.

        Parameters
        ----------

        field : {str, np.ndarray}, optional

        time : number, optional

        QUIVER : True

        vecx : 'ux'

        vecy : 'uy'

        nb_contours : 20

        type_plot : 'contourf'

        vmin : None

        vmax : None

        cmap : None (usually 'viridis')

        numfig : None

        """

        if equation is not None:
            self.set_equation_crosssection(equation)
        equation = self._equation

        is_field_ready = False

        self._has_uxuy = self.sim.state.has_vars("vx", "vy")

        key_field = None
        if field is None:
            key_field = self.get_key_field_to_plot(
                forbid_compute=time is not None
            )
        elif isinstance(field, np.ndarray):
            key_field = "given array"
            is_field_ready = True
        elif isinstance(field, str):
            key_field = field

        assert key_field is not None

        xlabel, ylabel = _get_xylabels_from_equation(equation)
        vecx = vector + xlabel
        vecy = vector + ylabel

        keys_state_phys = self.sim.state.keys_state_phys
        if vecx not in keys_state_phys or vecy not in keys_state_phys:
            QUIVER = False

        if (
            time is None
            and not is_field_ready
            and not self.sim.params.ONLY_COARSE_OPER
        ):
            # we have to get the field from the state
            time = self.sim.time_stepping.t
        else:
            # we have to get the field from a file
            self.set_of_phys_files.update_times()
            if key_field not in self.sim.state.keys_state_phys:
                error_message = (
                    f'key "{key_field}" not in state.keys_state_phys '
                    f"({self.sim.state.keys_state_phys})."
                )

                if time is not None:
                    error_message += (
                        "\nThe quantity cannot be computed because "
                        "time is not None."
                    )
                elif key_field in self.sim.state.keys_computable:
                    if self.sim.params.ONLY_COARSE_OPER:
                        error_message += (
                            f'\n"{key_field}" in sim.state.keys_computable '
                            "but sim.params.ONLY_COARSE_OPER is True"
                        )
                else:
                    error_message += (
                        f"\nThe quantity cannot be computed because "
                        '"{key_field}" in sim.state.keys_computable.'
                    )

                raise ValueError(error_message)

        if not is_field_ready:
            field, time = self.get_field_to_plot(
                key=key_field, time=time, equation=equation
            )
            if QUIVER:
                vecx, time = self.get_field_to_plot(
                    key=vecx, time=time, equation=equation
                )
                vecy, time = self.get_field_to_plot(
                    key=vecy, time=time, equation=equation
                )

        if mpi.rank == 0:
            if numfig is None:
                fig, ax = self.output.figure_axe()
            else:
                fig, ax = self.output.figure_axe(numfig=numfig)

            x_seq, y_seq = self._get_grid1d(equation)

            [XX_seq, YY_seq] = np.meshgrid(x_seq, y_seq)
            try:
                cmap = plt.get_cmap(cmap)
            except ValueError:
                print(
                    "Use matplotlib >= 1.5.0 for new standard colorschemes.\
                       Installed matplotlib :"
                    + plt.matplotlib.__version__
                )
                cmap = plt.get_cmap("jet")

            if type_plot == "contourf":
                contours = ax.contourf(
                    x_seq,
                    y_seq,
                    field,
                    nb_contours,
                    vmin=vmin,
                    vmax=vmax,
                    cmap=cmap,
                )
                fig.colorbar(contours)
                fig.contours = contours
            elif type_plot == "pcolor":
                pc = ax.pcolormesh(
                    x_seq,
                    y_seq,
                    field,
                    shading="nearest",
                    vmin=vmin,
                    vmax=vmax,
                    cmap=cmap,
                )
                fig.colorbar(pc)
        else:
            ax = None

        if QUIVER:
            quiver, vmax = self._quiver_plot(ax, vecx, vecy)
        else:
            vmax = None

        if mpi.rank == 0:
            ax.set_xlabel(xlabel)
            ax.set_ylabel(ylabel)
            self._set_title(ax, key_field, time, vmax)

            if SCALED:
                ax.axis("scaled")

            fig.tight_layout()
            fig.canvas.draw()
            plt.pause(1e-3)

    def plot_mean(
        self,
        field=None,
        tmin=None,
        tmax=None,
        QUIVER=True,
        vector="v",
        equation=None,
        nb_contours=20,
        type_plot="contourf",
        vmin=None,
        vmax=None,
        cmap=None,
        numfig=None,
        SCALED=True,
    ):
        """Plot the time average of a field.

        Parameters
        ----------

        field : str, optional

        tmin : number, optional

        tmax : number, optional

        QUIVER : True

        vecx : 'ux'

        vecy : 'uy'

        nb_contours : 20

        type_plot : 'contourf'

        vmin : None

        vmax : None

        cmap : None (usually 'viridis')

        numfig : None

        SCALED : True

        """

        if equation is not None:
            self.set_equation_crosssection(equation)
        equation = self._equation

        self._has_uxuy = self.sim.state.has_vars("vx", "vy")

        key_field = None
        if field is None:
            key_field = self.get_key_field_to_plot(forbid_compute=True)
        elif isinstance(field, str):
            key_field = field

        assert key_field is not None

        xlabel, ylabel = _get_xylabels_from_equation(equation)
        vecx = vector + xlabel
        vecy = vector + ylabel

        keys_state_phys = self.sim.state.keys_state_phys
        if vecx not in keys_state_phys or vecy not in keys_state_phys:
            QUIVER = False

        self.set_of_phys_files.update_times()
        times = self.set_of_phys_files.times
        if tmin is None:
            # get tmin from times
            tmin = times.min()

        if tmax is None:
            # get tmax from times
            tmax = times.max()

        # we have to get the field from a file
        if key_field not in self.sim.state.keys_state_phys:
            error_message = (
                f'key "{key_field}" not in state.keys_state_phys '
                f"({self.sim.state.keys_state_phys})."
            )
            raise ValueError(error_message)

        # get times for average
        times_avg = times[(times >= tmin) & (times <= tmax)]

        # initialize
        time = times_avg[0]
        field_onetime, time = self.get_field_to_plot(
            key=key_field, time=time, equation=equation
        )
        field_avg = np.zeros_like(field_onetime)
        vecx_avg = np.zeros_like(field_onetime)
        vecy_avg = np.zeros_like(field_onetime)
        # time average
        for time in times_avg:
            field_onetime, time = self.get_field_to_plot(
                key=key_field, time=time, equation=equation
            )
            field_avg += field_onetime
            if QUIVER:
                field_onetime, time = self.get_field_to_plot(
                    key=vecx, time=time, equation=equation
                )
                vecx_avg += field_onetime
                field_onetime, time = self.get_field_to_plot(
                    key=vecy, time=time, equation=equation
                )
                vecy_avg += field_onetime
        field_avg /= times_avg.size
        vecx_avg /= times_avg.size
        vecy_avg /= times_avg.size

        # plot
        if mpi.rank == 0:
            if numfig is None:
                fig, ax = self.output.figure_axe()
            else:
                fig, ax = self.output.figure_axe(numfig=numfig)

            x_seq, y_seq = self._get_grid1d(equation)

            [XX_seq, YY_seq] = np.meshgrid(x_seq, y_seq)
            try:
                cmap = plt.get_cmap(cmap)
            except ValueError:
                print(
                    "Use matplotlib >= 1.5.0 for new standard colorschemes.\
                       Installed matplotlib :"
                    + plt.matplotlib.__version__
                )
                cmap = plt.get_cmap("jet")

            if type_plot == "contourf":
                contours = ax.contourf(
                    x_seq,
                    y_seq,
                    field_avg,
                    nb_contours,
                    vmin=vmin,
                    vmax=vmax,
                    cmap=cmap,
                )
                fig.colorbar(contours)
                fig.contours = contours
            elif type_plot == "pcolor":
                pc = ax.pcolormesh(
                    x_seq,
                    y_seq,
                    field_avg,
                    shading="nearest",
                    vmin=vmin,
                    vmax=vmax,
                    cmap=cmap,
                )
                fig.colorbar(pc)
        else:
            ax = None

        if QUIVER:
            quiver, vmax = self._quiver_plot(ax, vecx_avg, vecy_avg)
        else:
            vmax = None

        if mpi.rank == 0:
            ax.set_xlabel(xlabel)
            ax.set_ylabel(ylabel)
            title = key_field + f" $(tmin={tmin:8.6g}, tmax={tmax:8.6g})$"
            if vmax is not None:
                title += r", $|\vec{v}|_{max} = $" + f"{vmax:.3f}"
            ax.set_title(title + "\n" + self.output.summary_simul)

            if SCALED:
                ax.axis("scaled")

            fig.tight_layout()
            fig.canvas.draw()
            plt.pause(1e-3)

    def _quiver_plot(self, ax, vecx="ux", vecy="uy", XX=None, YY=None):
        """Superimposes a quiver plot of velocity vectors with a given axis
        object corresponding to a 2D contour plot.

        """
        if isinstance(vecx, str):
            vecx, time = self.get_field_to_plot(vecx)

        if isinstance(vecy, str):
            vecy, time = self.get_field_to_plot(vecy)

        if XX is None and YY is None:
            equation = self._equation

            x_seq, y_seq = self._get_grid1d(equation)

            [XX, YY] = np.meshgrid(x_seq, y_seq)

        if mpi.rank == 0:
            # local variable 'normalize_diff' is assigned to but never used
            # normalize_diff = (
            #     (np.max(np.sqrt(vecx**2 + vecy**2)) -
            #      np.min(np.sqrt(vecx**2 + vecy**2))) /
            #     np.max(np.sqrt(vecx**2 + vecy**2)))
            vmax = np.max(np.sqrt(vecx ** 2 + vecy ** 2))
            # Quiver is normalized by the vmax
            # copy to avoid a bug
            skip = self._skip_quiver
            vecx_c = vecx[::skip, ::skip].copy()
            vecy_c = vecy[::skip, ::skip].copy()
            quiver = ax.quiver(
                XX[::skip, ::skip],
                YY[::skip, ::skip],
                vecx_c / vmax,
                vecy_c / vmax,
            )
        else:
            quiver = vmax = None

        return quiver, vmax



"""State stratified NS3D solver (:mod:`fluidsim.solvers.ns3d.strat.state`)
==========================================================================
"""

import numpy as np

from ..state import StateNS3D


class StateNS3DStrat(StateNS3D):
    """Contains the variables corresponding to the state and handles the
    access to other fields for the solver NS3D.

    """

    @staticmethod
    def _complete_info_solver(info_solver):
        """Complete the ParamContainer info_solver.

        This is a static method!
        """
        keys_state_phys = ["vx", "vy", "vz", "b"]

        info_solver.classes.State._set_attribs(
            {
                "keys_state_spect": [k + "_fft" for k in keys_state_phys],
                "keys_state_phys": keys_state_phys,
                "keys_phys_needed": keys_state_phys,
                "keys_computable": ["rotz", "divh"],
                "keys_linear_eigenmodes": ["rot_fft"],
            }
        )

    def init_from_vxvyfft(self, vx_fft, vy_fft):
        self.state_spect.set_var("vx_fft", vx_fft)
        self.state_spect.set_var("vy_fft", vy_fft)
        self.state_spect.set_var("vz_fft", np.zeros_like(vx_fft))
        self.state_spect.set_var("b_fft", np.zeros_like(vx_fft))

        vx = self.oper.ifft3d(vx_fft)
        self.state_phys.set_var("vx", vx)
        self.state_phys.set_var("vy", self.oper.ifft3d(vy_fft))
        self.state_phys.set_var("vz", np.zeros_like(vx))
        self.state_phys.set_var("b", np.zeros_like(vx))

    def init_from_vxvyvzfft(self, vx_fft, vy_fft, vz_fft):
        self.state_spect.set_var("vx_fft", vx_fft)
        self.state_spect.set_var("vy_fft", vy_fft)
        self.state_spect.set_var("vz_fft", vz_fft)
        self.state_spect.set_var("b_fft", np.zeros_like(vx_fft))

        self.statephys_from_statespect()
        self.statespect_from_statephys()



"""Operators 1d (:mod:`fluidsim.operators.operators1d`)
=======================================================

Provides

.. autoclass:: OperatorsPseudoSpectral1D
   :members:
   :private-members:

"""

import numpy as np
from fluiddyn.calcul.easypyfft import FFTW1DReal2Complex
from .base import OperatorsBase1D
from ..base.setofvariables import SetOfVariables


class OperatorsPseudoSpectral1D(OperatorsBase1D):
    """1D operators for pseudospectral solvers."""

    @staticmethod
    def _complete_params_with_default(params):
        """This static method is used to complete the *params* container."""
        params = OperatorsBase1D._complete_params_with_default(params)
        params.oper._set_attribs(
            {"type_fft": "sequential", "coef_dealiasing": 2.0 / 3}
        )
        return params

    def __init__(self, params, SEQUENTIAL=None):
        super().__init__(params)

        assert params.oper.type_fft == "sequential"
        nx = params.oper.nx
        opfft = FFTW1DReal2Complex(nx)
        self.fft = opfft.fft
        self.ifft = opfft.ifft
        self.fft_as_arg = opfft.fft_as_arg
        self.ifft_as_arg = opfft.ifft_as_arg
        self.sum_wavenumbers = opfft.sum_wavenumbers

        self.shapeK = self.shapeK_loc = opfft.shapeK
        self.deltakx = 2 * np.pi / self.Lx
        self.nkx = self.shapeK[0]
        self.kx = self.deltakx * np.arange(self.nkx)
        assert self.nkx == self.nx // 2 + 1

        self.coef_dealiasing = params.oper.coef_dealiasing
        kx_max = self.deltakx * (self.nkx - 1)
        assert kx_max == self.kx.max()

        CONDKX = abs(self.kx) > self.coef_dealiasing * kx_max
        self.where_dealiased = np.array(CONDKX, dtype=bool)

        # for spectra
        self.nkxE = self.nx // 2 + 1
        self.nkxE2 = (self.nx + 1) // 2

        # print('nkxE, nkxE2', self.nkxE, self.nkxE2)

        self.kxE = self.deltakx * np.arange(self.nkxE)
        self.khE = self.kxE
        self.nkhE = self.nkxE

        self.KX = self.kx
        self.K2 = self.KX ** 2
        self.K4 = self.K2 ** 2
        self.K8 = self.K4 ** 2
        self.K = self.KX

    def produce_long_str_describing_oper(self):
        return super().produce_long_str_describing_oper("Pseudospectral")

    def dealiasing(self, *args):
        for thing in args:
            if isinstance(thing, SetOfVariables):
                self.dealiasing_setofvar(thing)
            elif isinstance(thing, np.ndarray):
                self.dealiasing_variable(thing)

    def dealiasing_variable(self, f_fft):
        f_fft[self.where_dealiased] = 0.0

    def dealiasing_setofvar(self, sov):
        for ik in range(sov.nvar):
            self.dealiasing_variable(sov[ik])

    def pxffft_from_fft(self, f_fft):
        """Return the first derivative of f_fft in spectral space."""
        kx = self.kx
        px_f_fft = 1j * kx * f_fft
        return px_f_fft

    def get_phases_random(self):
        alpha = np.random.uniform(-0.5, 0.5)
        beta = alpha + 0.5 if alpha < 0 else alpha - 0.5

        phase_alpha = alpha * self.deltax * self.kx
        phase_beta = beta * self.deltax * self.kx
        return phase_alpha, phase_beta



"""State of the variables (:mod:`fluidsim.base.dedalus.state`)
===============================================================

Provides:

.. autoclass:: StateDedalus
   :members:
   :private-members:

"""
import numpy as np

from ..state import StateBase


class StatePhysDedalus:
    def __init__(self, dedalus_solver, keys):
        self.dedalus_solver = dedalus_solver
        self.keys = keys

    def get_var(self, key):
        if self.dedalus_solver is None:
            return np.zeros(1)

        field = self.dedalus_solver.state[key]
        field.set_scales(1.0)
        return field["g"].transpose()

    def set_var(self, key, value):
        field = self.dedalus_solver.state[key]
        field.set_scales(1.0)
        field["g"] = value.transpose()

    def initialize(self, value):
        for key in self.keys:
            self.get_var(key).fill(value)

    @property
    def nbytes(self):
        return 1

    @property
    def info(self):
        return "Dedalus state"


class StateDedalus(StateBase):
    @staticmethod
    def _complete_info_solver(info_solver):
        """Complete the ParamContainer info_solver.

        This is a static method!
        """
        info_solver.classes.State._set_attribs(
            {
                "keys_state_phys": ["b", "vx", "vz"],
                "keys_computable": ["dz_b", "dz_vx", "dz_vz"],
                "keys_phys_needed": ["b", "vx", "vz"],
            }
        )

    def compute(self, key):
        """Compute scalar fields such a component of the velocity or vorticity."""
        return self.state_phys.get_var(key)

    def __init__(self, sim, oper=None):
        super().__init__(sim, oper)
        sim.init_dedalus()
        if self.sim.params.ONLY_COARSE_OPER:
            solver = None
        else:
            solver = self.sim.dedalus_solver

        self.state_phys = StatePhysDedalus(solver, self.keys_state_phys)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

import numpy as np

import matplotlib.pyplot as plt


f = 1.0
c = 1.0

L_D = c / f

Lx = 30.0
nx = 64
nkx = nx / 2.0 + 1

kx = 2 * np.pi / Lx * np.arange(nkx)

omega_theo = np.sqrt(f ** 2 + (c * kx) ** 2)

# numerical results
k_num = np.array([0.63, 1.88, 2.51, 4.19])

frequency = np.array([0.1875, 0.34, 0.43, 0.7])


ikx = 14


fig = plt.figure(2)
fig.clf()
plt.hold(True)

plt.plot(kx * L_D, omega_theo / (2 * np.pi), "k", linewidth=2)

plt.plot(kx * L_D, kx * L_D / (2 * np.pi), "k--", linewidth=0.5)

plt.plot(kx[ikx] * L_D, omega_theo[ikx] / (2 * np.pi), "rx", linewidth=2)

plt.plot(k_num * L_D, frequency, "bo", linewidth=2)


plt.xlabel("k L_D")
plt.ylabel("frequency")
plt.show()



"""Energy budget (:mod:`fluidsim.solvers.ns2d.output.spect_energy_budget`)
==========================================================================

.. autoclass:: SpectralEnergyBudgetNS2D
   :members:
   :private-members:

"""

import numpy as np
import h5py


from fluidsim.base.output.spect_energy_budget import (
    SpectralEnergyBudgetBase,
    cumsum_inv,
)


class SpectralEnergyBudgetNS2D(SpectralEnergyBudgetBase):
    """Save and plot energy budget in spectral space."""

    def compute(self):
        """compute the spectral energy budget at one time."""
        oper = self.sim.oper

        ux = self.sim.state.state_phys.get_var("ux")
        uy = self.sim.state.state_phys.get_var("uy")

        rot_fft = self.sim.state.state_spect.get_var("rot_fft")
        ux_fft, uy_fft = oper.vecfft_from_rotfft(rot_fft)

        px_rot_fft, py_rot_fft = oper.gradfft_from_fft(rot_fft)
        px_rot = oper.ifft2(px_rot_fft)
        py_rot = oper.ifft2(py_rot_fft)

        px_ux_fft, py_ux_fft = oper.gradfft_from_fft(ux_fft)
        px_ux = oper.ifft2(px_ux_fft)
        py_ux = oper.ifft2(py_ux_fft)

        px_uy_fft, py_uy_fft = oper.gradfft_from_fft(uy_fft)
        px_uy = oper.ifft2(px_uy_fft)
        py_uy = oper.ifft2(py_uy_fft)

        Frot = -ux * px_rot - uy * (py_rot + self.params.beta)
        Frot_fft = oper.fft2(Frot)
        oper.dealiasing(Frot_fft)

        Fx = -ux * px_ux - uy * (py_ux)
        Fx_fft = oper.fft2(Fx)
        oper.dealiasing(Fx_fft)

        Fy = -ux * px_uy - uy * (py_uy)
        Fy_fft = oper.fft2(Fy)
        oper.dealiasing(Fy_fft)

        transferZ_fft = (
            np.real(rot_fft.conj() * Frot_fft + rot_fft * Frot_fft.conj()) / 2.0
        )
        # print ('sum(transferZ) = {0:9.4e} ; sum(abs(transferZ)) = {1:9.4e}'
        #       ).format(self.sum_wavenumbers(transferZ_fft),
        #                self.sum_wavenumbers(abs(transferZ_fft)))

        transferE_fft = (
            np.real(
                ux_fft.conj() * Fx_fft
                + ux_fft * Fx_fft.conj()
                + uy_fft.conj() * Fy_fft
                + uy_fft * Fy_fft.conj()
            )
            / 2.0
        )
        # print ('sum(transferE) = {0:9.4e} ; sum(abs(transferE)) = {1:9.4e}'
        #       ).format(self.sum_wavenumbers(transferE_fft),
        #                self.sum_wavenumbers(abs(transferE_fft)))

        transfer2D_E = self.spectrum2D_from_fft(transferE_fft)
        transfer2D_Z = self.spectrum2D_from_fft(transferZ_fft)

        dict_results = {
            "transfer2D_E": transfer2D_E,
            "transfer2D_Z": transfer2D_Z,
        }
        return dict_results

    def _online_plot_saving(self, dict_results):
        transfer2D_E = dict_results["transfer2D_E"]
        transfer2D_Z = dict_results["transfer2D_Z"]
        khE = self.oper.khE
        PiE = cumsum_inv(transfer2D_E) * self.oper.deltak
        PiZ = cumsum_inv(transfer2D_Z) * self.oper.deltak
        self.axe_a.plot(khE + khE[1], PiE, "k")
        self.axe_b.plot(khE + khE[1], PiZ, "g")

    def plot(self, tmin=0, tmax=1000, delta_t=2):

        with h5py.File(self.path_file, "r") as h5file:
            dset_times = h5file["times"]
            dset_khE = h5file["khE"]
            khE = dset_khE[...]
            khE = khE + khE[1]

            dset_transferE = h5file["transfer2D_E"]
            dset_transferZ = h5file["transfer2D_Z"]

            # nb_spectra = dset_times.shape[0]
            times = dset_times[...]
            # nt = len(times)

            delta_t_save = np.mean(times[1:] - times[0:-1])
            delta_i_plot = int(np.round(delta_t / delta_t_save))

            if delta_i_plot == 0 and delta_t != 0.0:
                delta_i_plot = 1
            delta_t = delta_i_plot * delta_t_save

            imin_plot = np.argmin(abs(times - tmin))
            imax_plot = np.argmin(abs(times - tmax))

            to_print = "plot(tmin={}, tmax={}, delta_t={:.2f})".format(
                tmin, tmax, delta_t
            )
            print(to_print)

            tmin_plot = times[imin_plot]
            tmax_plot = times[imax_plot]
            print(
                """plot spectral energy budget
    tmin = {:8.6g} ; tmax = {:8.6g} ; delta_t = {:8.6g}
    imin = {:8d} ; imax = {:8d} ; delta_i = {:8d}""".format(
                    tmin_plot,
                    tmax_plot,
                    delta_t,
                    imin_plot,
                    imax_plot,
                    delta_i_plot,
                )
            )

            fig, ax1 = self.output.figure_axe()
            ax1.set_xlabel("$k_h$")
            ax1.set_ylabel("spectra")
            ax1.set_xscale("log")
            ax1.set_yscale("linear")

            if delta_t != 0.0:
                for it in range(imin_plot, imax_plot, delta_i_plot):

                    transferE = dset_transferE[it]
                    transferZ = dset_transferZ[it]

                    PiE = cumsum_inv(transferE) * self.oper.deltak
                    PiZ = cumsum_inv(transferZ) * self.oper.deltak

                    ax1.plot(khE, PiE, "k", linewidth=1)
                    ax1.plot(khE, PiZ, "g", linewidth=1)

            transferE = dset_transferE[imin_plot:imax_plot].mean(0)
            transferZ = dset_transferZ[imin_plot:imax_plot].mean(0)

        PiE = cumsum_inv(transferE) * self.oper.deltak
        PiZ = cumsum_inv(transferZ) * self.oper.deltak

        ax1.plot(khE, PiE, "r", linewidth=2)
        ax1.plot(khE, PiZ, "m", linewidth=2)



import numpy as np
import h5py

from fluiddyn.util import mpi

from fluidsim.base.output.spect_energy_budget import (
    SpectralEnergyBudgetBase,
    cumsum_inv,
    inner_prod,
)


class SpectralEnergyBudgetSW1LWaves(SpectralEnergyBudgetBase):
    def __init__(self, output):

        params = output.sim.params
        self.c2 = params.c2
        self.f = params.f

        super().__init__(output)

    def compute(self):
        """compute spectral energy budget the one time."""
        oper = self.sim.oper

        # print_memory_usage('start function compute seb')

        ux = self.sim.state.state_phys.get_var("ux")
        uy = self.sim.state.state_phys.get_var("uy")
        eta = self.sim.state.state_phys.get_var("eta")
        h = 1.0 + eta

        Jx = h * ux
        Jy = h * uy
        Jx_fft = oper.fft2(Jx)
        Jy_fft = oper.fft2(Jy)
        del (Jx, Jy)

        get_var = self.sim.state.get_var
        ux_fft = get_var("ux_fft")
        uy_fft = get_var("uy_fft")
        eta_fft = get_var("eta_fft")
        h_fft = eta_fft.copy()
        if mpi.rank == 0:
            h_fft[0, 0] = 1.0

        rot_fft = oper.rotfft_from_vecfft(ux_fft, uy_fft)
        urx_fft, ury_fft = oper.vecfft_from_rotfft(rot_fft)
        del rot_fft
        urx = oper.ifft2(urx_fft)
        ury = oper.ifft2(ury_fft)

        q_fft, div_fft, a_fft = self.oper.qdafft_from_uxuyetafft(
            ux_fft, uy_fft, eta_fft
        )

        # if self.params.f != 0:
        #     udx_fft, udy_fft = oper.vecfft_from_divfft(div_fft)
        #     ugx_fft, ugy_fft, etag_fft = \
        #         self.oper.uxuyetafft_from_qfft(q_fft)
        #     uax_fft, uay_fft, etaa_fft = \
        #         self.oper.uxuyetafft_from_afft(a_fft)
        #     del(a_fft)
        #     # velocity influenced by linear terms
        #     u_infl_lin_x = udx_fft + uax_fft
        #     u_infl_lin_y = udy_fft + uay_fft

        udx_fft, udy_fft = oper.vecfft_from_divfft(div_fft)
        udx = oper.ifft2(udx_fft)
        udy = oper.ifft2(udy_fft)
        div = oper.ifft2(div_fft)
        del div_fft

        # print_memory_usage('before starting computing fluxes')

        # compute flux of Charney PE
        Fq_fft = self.fnonlinfft_from_uxuy_funcfft(urx, ury, q_fft)
        transferCPE_fft = inner_prod(q_fft, Fq_fft)
        del (q_fft, Fq_fft)
        transfer2D_CPE = self.spectrum2D_from_fft(transferCPE_fft)
        del transferCPE_fft

        #         print(
        # ('sum(transfer2D_CPE) = {0:9.4e} ; sum(abs(transfer2D_CPE)) = {1:9.4e}'
        # ).format(
        # np.sum(transfer2D_CPE),
        # np.sum(abs(transfer2D_CPE)))
        # )

        px_h_fft, py_h_fft = oper.gradfft_from_fft(eta_fft)
        px_h = oper.ifft2(px_h_fft)
        py_h = oper.ifft2(py_h_fft)

        F_rh = -urx * px_h - ury * py_h
        F_dh = -udx * px_h - udy * py_h - h * div / 2
        F_de = -udx * px_h - udy * py_h - eta * div / 2
        del (px_h, py_h)
        F_rh_fft = oper.fft2(F_rh)
        F_dh_fft = oper.fft2(F_dh)
        F_de_fft = oper.fft2(F_de)
        del (F_rh, F_dh, F_de)
        oper.dealiasing(F_rh_fft)
        oper.dealiasing(F_dh_fft)
        oper.dealiasing(F_de_fft)

        transferEAr_fft = self.c2 * inner_prod(h_fft, F_rh_fft)
        transferEPd_fft = self.c2 * inner_prod(h_fft, F_dh_fft)
        transferEAd_fft = self.c2 * inner_prod(eta_fft, F_de_fft)
        del (F_rh_fft, F_dh_fft, F_de_fft)

        transfer2D_EAr = self.spectrum2D_from_fft(transferEAr_fft)
        transfer2D_EPd = self.spectrum2D_from_fft(transferEPd_fft)
        transfer2D_EAd = self.spectrum2D_from_fft(transferEAd_fft)
        del (transferEAr_fft, transferEPd_fft, transferEAd_fft)

        # print_memory_usage('after transfer2D_EAr')

        #         print(
        # ('sum(transfer2D_EAr) = {0:9.4e} ; sum(abs(transfer2D_EAr)) = {1:9.4e}'
        # ).format(
        # np.sum(transfer2D_EAr),
        # np.sum(abs(transfer2D_EAr)))
        # )

        #         print(
        # ('sum(transfer2D_EAd) = {0:9.4e} ; sum(abs(transfer2D_EAd)) = {1:9.4e}'
        # ).format(
        # np.sum(transfer2D_EAd),
        # np.sum(abs(transfer2D_EAd)))
        # )

        hdiv_fft = oper.fft2(h * div)
        convP_fft = self.c2 / 2.0 * inner_prod(h_fft, hdiv_fft)
        convP2D = self.spectrum2D_from_fft(convP_fft)
        del (convP_fft, h_fft, hdiv_fft)

        EP = self.c2 / 2 * h * h
        EP_fft = oper.fft2(EP)
        del (EP, h)
        px_EP_fft, py_EP_fft = oper.gradfft_from_fft(EP_fft)
        del EP_fft

        convK_fft = (
            1.0
            / 2
            * (
                -inner_prod(ux_fft, px_EP_fft)
                - inner_prod(uy_fft, py_EP_fft)
                - self.c2 * inner_prod(Jx_fft, px_h_fft)
                - self.c2 * inner_prod(Jy_fft, py_h_fft)
            )
        )
        del (px_h_fft, py_h_fft, px_EP_fft, py_EP_fft)
        convK2D = self.spectrum2D_from_fft(convK_fft)
        del convK_fft

        # print_memory_usage('after convK2D')

        #         print(
        # ('sum(convP2D-convK2D)*deltak = {0:9.4e}, sum(convP2D)*deltak = {1:9.4e}'
        # ).format(
        # np.sum(convP2D-convK2D)*self.oper.deltak,
        # np.sum(convP2D)*self.oper.deltak
        # )
        # )

        #         print(
        # ('                                           sum(convK2D)*deltak = {0:9.4e}'
        # ).format(
        # np.sum(convK2D)*self.oper.deltak
        # )
        # )

        Fxrd_fft, Fxdd_fft = self.fnonlinfft_from_uruddivfunc(
            urx, ury, udx, udy, div, udx_fft, udx
        )
        Fyrd_fft, Fydd_fft = self.fnonlinfft_from_uruddivfunc(
            urx, ury, udx, udy, div, udy_fft, udy
        )
        Fxrr_fft, Fxdr_fft = self.fnonlinfft_from_uruddivfunc(
            urx, ury, udx, udy, div, urx_fft, urx
        )
        Fyrr_fft, Fydr_fft = self.fnonlinfft_from_uruddivfunc(
            urx, ury, udx, udy, div, ury_fft, ury
        )

        transferErrr_fft = inner_prod(urx_fft, Fxrr_fft) + inner_prod(
            ury_fft, Fyrr_fft
        )
        transfer2D_Errr = self.spectrum2D_from_fft(transferErrr_fft)
        del transferErrr_fft
        #         print(
        # ('sum(transfer2D_Errr) = {0:9.4e} ; sum(abs(transfer2D_Errr)) = {1:9.4e}'
        # ).format(
        # np.sum(transfer2D_Errr),
        # np.sum(abs(transfer2D_Errr)))
        # )

        transferEdrd_fft = inner_prod(udx_fft, Fxrd_fft) + inner_prod(
            udy_fft, Fyrd_fft
        )
        transfer2D_Edrd = self.spectrum2D_from_fft(transferEdrd_fft)
        del transferEdrd_fft
        #         print(
        # ('sum(transfer2D_Edrd) = {0:9.4e} ; sum(abs(transfer2D_Edrd)) = {1:9.4e}'
        # ).format(
        # np.sum(transfer2D_Edrd),
        # np.sum(abs(transfer2D_Edrd)))
        # )
        Clfromqq = inner_prod(udx_fft, Fxrr_fft) + inner_prod(udy_fft, Fyrr_fft)
        transferEdrr_rrd_fft = (
            Clfromqq
            + inner_prod(urx_fft, Fxrd_fft)
            + inner_prod(ury_fft, Fyrd_fft)
        )
        Clfromqq = self.spectrum2D_from_fft(Clfromqq)
        transfer2D_Edrr_rrd = self.spectrum2D_from_fft(transferEdrr_rrd_fft)
        del transferEdrr_rrd_fft
        #         print(
        # ('sum(transfer2D_Edrr_rrd) = {0:9.4e} ; '
        # 'sum(abs(transfer2D_Edrr_rrd)) = {1:9.4e}'
        # ).format(
        # np.sum(transfer2D_Edrr_rrd),
        # np.sum(abs(transfer2D_Edrr_rrd)))
        # )

        transferErdr_fft = inner_prod(urx_fft, Fxdr_fft) + inner_prod(
            ury_fft, Fydr_fft
        )
        transfer2D_Erdr = self.spectrum2D_from_fft(transferErdr_fft)
        del transferErdr_fft
        #         print(
        # ('sum(transfer2D_Erdr) = {0:9.4e} ; sum(abs(transfer2D_Erdr)) = {1:9.4e}'
        # ).format(
        # np.sum(transfer2D_Erdr),
        # np.sum(abs(transfer2D_Erdr)))
        # )

        transferEddd_fft = inner_prod(udx_fft, Fxdd_fft) + inner_prod(
            udy_fft, Fydd_fft
        )
        transfer2D_Eddd = self.spectrum2D_from_fft(transferEddd_fft)
        del transferEddd_fft
        #         print(
        # ('sum(transfer2D_Eddd) = {0:9.4e} ; sum(abs(transfer2D_Eddd)) = {1:9.4e}'
        # ).format(
        # np.sum(transfer2D_Eddd),
        # np.sum(abs(transfer2D_Eddd)))
        # )

        Cqfromll = inner_prod(urx_fft, Fxdd_fft) + inner_prod(ury_fft, Fydd_fft)
        transferEddr_rdd_fft = (
            Cqfromll
            + inner_prod(udx_fft, Fxdr_fft)
            + inner_prod(udy_fft, Fydr_fft)
        )

        Cqfromll = self.spectrum2D_from_fft(Cqfromll)
        transfer2D_Eddr_rdd = self.spectrum2D_from_fft(transferEddr_rdd_fft)
        del transferEddr_rdd_fft
        #         print(
        # ('sum(transfer2D_Eddr_rdd) = {0:9.4e} ; '
        # 'sum(abs(transfer2D_Eddr_rdd)) = {1:9.4e}'
        # ).format(
        # np.sum(transfer2D_Eddr_rdd),
        # np.sum(abs(transfer2D_Eddr_rdd)))
        # )

        Fx_ru_fft = Fxrr_fft + Fxrd_fft
        del (Fxrr_fft, Fxrd_fft)
        Fy_ru_fft = Fyrr_fft + Fyrd_fft
        del (Fyrr_fft, Fyrd_fft)
        Fx_du_fft = Fxdr_fft + Fxdd_fft
        del (Fxdr_fft, Fxdd_fft)
        Fy_du_fft = Fydr_fft + Fydd_fft
        del (Fydr_fft, Fydd_fft)

        # print_memory_usage('after Fy_du_fft')

        etaux = eta * ux
        etauy = eta * uy

        etaux_fft = oper.fft2(etaux)
        etauy_fft = oper.fft2(etauy)
        del (etaux, etauy)

        px_etaux_fft, py_etaux_fft = oper.gradfft_from_fft(etaux_fft)
        px_etauy_fft, py_etauy_fft = oper.gradfft_from_fft(etauy_fft)

        px_etaux = oper.ifft2(px_etaux_fft)
        del px_etaux_fft
        py_etaux = oper.ifft2(py_etaux_fft)
        del py_etaux_fft
        px_etauy = oper.ifft2(px_etauy_fft)
        del px_etauy_fft
        py_etauy = oper.ifft2(py_etauy_fft)
        del py_etauy_fft

        Fx_reu = -urx * px_etaux - ury * py_etaux
        Fx_reu_fft = oper.fft2(Fx_reu)
        del Fx_reu

        Fx_deu = -udx * px_etaux - udy * py_etaux - 0.5 * div * eta * ux
        del (px_etaux, py_etaux)
        Fx_deu_fft = oper.fft2(Fx_deu)
        del Fx_deu

        Fy_reu = -urx * px_etauy - ury * py_etauy
        Fy_reu_fft = oper.fft2(Fy_reu)
        del Fy_reu

        Fy_deu = -udx * px_etauy - udy * py_etauy - 0.5 * div * eta * uy
        del (px_etauy, py_etauy)
        Fy_deu_fft = oper.fft2(Fy_deu)
        del Fy_deu

        transferEureu_fft = 0.5 * (
            inner_prod(ux_fft, Fx_reu_fft)
            + inner_prod(uy_fft, Fy_reu_fft)
            + inner_prod(etaux_fft, Fx_ru_fft)
            + inner_prod(etauy_fft, Fy_ru_fft)
        )
        del (Fx_reu_fft, Fy_reu_fft, Fx_ru_fft, Fy_ru_fft)

        transfer2D_Eureu = self.spectrum2D_from_fft(transferEureu_fft)
        #         print(
        # ('sum(transferEureu_fft) = {0:9.4e} ; '
        # 'sum(abs(transferEureu_fft)) = {1:9.4e}'
        # ).format(
        # np.sum(transfer2D_Eureu),
        # np.sum(abs(transfer2D_Eureu)))
        # )

        transferEudeu_fft = 0.5 * (
            inner_prod(ux_fft, Fx_deu_fft)
            + inner_prod(uy_fft, Fy_deu_fft)
            + inner_prod(etaux_fft, Fx_du_fft)
            + inner_prod(etauy_fft, Fy_du_fft)
        )
        del (Fx_deu_fft, Fy_deu_fft, Fx_du_fft, Fy_du_fft)
        del (etaux_fft, etauy_fft)

        transfer2D_Eudeu = self.spectrum2D_from_fft(transferEudeu_fft)
        del transferEudeu_fft
        #         print(
        # ('sum(transferEudeu_fft) = {0:9.4e} ; '
        # 'sum(abs(transferEudeu_fft)) = {1:9.4e}'
        # ).format(
        # np.sum(transfer2D_Eudeu),
        # np.sum(abs(transfer2D_Eudeu)))
        # )

        transfer2D_EKr = (
            transfer2D_Errr
            + transfer2D_Edrd
            + transfer2D_Edrr_rrd
            + transfer2D_Eureu
        )
        transfer2D_EKd = (
            transfer2D_Erdr
            + transfer2D_Eddd
            + transfer2D_Eddr_rdd
            + transfer2D_Eudeu
        )

        # print_memory_usage('end of function compute seb')

        dict_results = {
            "transfer2D_EAr": transfer2D_EAr,
            "transfer2D_EAd": transfer2D_EAd,
            "transfer2D_EPd": transfer2D_EPd,
            "transfer2D_EKr": transfer2D_EKr,
            "transfer2D_EKd": transfer2D_EKd,
            "transfer2D_Errr": transfer2D_Errr,
            "transfer2D_Edrd": transfer2D_Edrd,
            "Clfromqq": Clfromqq,
            "transfer2D_Edrr_rrd": transfer2D_Edrr_rrd,
            "transfer2D_Erdr": transfer2D_Erdr,
            "transfer2D_Eddd": transfer2D_Eddd,
            "Cqfromll": Cqfromll,
            "transfer2D_Eddr_rdd": transfer2D_Eddr_rdd,
            "transfer2D_Eureu": transfer2D_Eureu,
            "transfer2D_Eudeu": transfer2D_Eudeu,
            "convP2D": convP2D,
            "convK2D": convK2D,
            "transfer2D_CPE": transfer2D_CPE,
        }
        return dict_results

    def _online_plot_saving(self, dict_results):

        transfer2D_CPE = dict_results["transfer2D_CPE"]
        transfer2D_EKr = dict_results["transfer2D_EKr"]
        transfer2D_EKd = dict_results["transfer2D_EKd"]
        transfer2D_EK = transfer2D_EKr + transfer2D_EKd
        transfer2D_EAr = dict_results["transfer2D_EAr"]
        transfer2D_EAd = dict_results["transfer2D_EAd"]
        transfer2D_EA = transfer2D_EAr + transfer2D_EAd
        convP2D = dict_results["convP2D"]
        convK2D = dict_results["convK2D"]
        khE = self.oper.khE
        PiCPE = cumsum_inv(transfer2D_CPE) * self.oper.deltak
        PiEK = cumsum_inv(transfer2D_EK) * self.oper.deltak
        PiEA = cumsum_inv(transfer2D_EA) * self.oper.deltak
        PiEtot = PiEK + PiEA
        CCP = cumsum_inv(convP2D) * self.oper.deltak
        CCK = cumsum_inv(convK2D) * self.oper.deltak

        self.axe_a.plot(khE + khE[1], PiEK, "r")
        self.axe_a.plot(khE + khE[1], PiEA, "b")
        self.axe_a.plot(khE + khE[1], PiEtot, "k", linewidth=2)
        self.axe_a.plot(khE + khE[1], CCP, "y")
        self.axe_a.plot(khE + khE[1], CCK, "y--")

        self.axe_b.plot(khE + khE[1], PiCPE, "g")

    def plot(self, tmin=0, tmax=1000, delta_t=2):

        with h5py.File(self.path_file, "r") as h5file:

            dset_times = h5file["times"]
            times = dset_times[...]
            # nt = len(times)

            dset_khE = h5file["khE"]
            khE = dset_khE[...]

            dset_transfer2D_EKr = h5file["transfer2D_EKr"]
            dset_transfer2D_EKd = h5file["transfer2D_EKd"]
            dset_transfer2D_EAr = h5file["transfer2D_EAr"]
            dset_transfer2D_EAd = h5file["transfer2D_EAd"]
            dset_transfer2D_EPd = h5file["transfer2D_EPd"]
            dset_convP2D = h5file["convP2D"]
            dset_convK2D = h5file["convK2D"]
            dset_transfer2D_CPE = h5file["transfer2D_CPE"]

            delta_t_save = np.mean(times[1:] - times[0:-1])
            delta_i_plot = int(np.round(delta_t / delta_t_save))
            if delta_i_plot == 0 and delta_t != 0.0:
                delta_i_plot = 1
            delta_t = delta_i_plot * delta_t_save

            imin_plot = np.argmin(abs(times - tmin))
            imax_plot = np.argmin(abs(times - tmax))

            tmin_plot = times[imin_plot]
            tmax_plot = times[imax_plot]

            to_print = "plot(tmin={}, tmax={}, delta_t={:.2f})".format(
                tmin, tmax, delta_t
            )
            print(to_print)

            to_print = """plot fluxes 2D
    tmin = {:8.6g} ; tmax = {:8.6g} ; delta_t = {:8.6g}
    imin = {:8d} ; imax = {:8d} ; delta_i = {:8d}""".format(
                tmin_plot, tmax_plot, delta_t, imin_plot, imax_plot, delta_i_plot
            )
            print(to_print)

            x_left_axe = 0.12
            z_bottom_axe = 0.36
            width_axe = 0.85
            height_axe = 0.57

            size_axe = [x_left_axe, z_bottom_axe, width_axe, height_axe]
            fig, ax1 = self.output.figure_axe(size_axe=size_axe)
            ax1.set_xlabel("$k_h$")
            ax1.set_ylabel("transfers")
            ax1.set_title("energy flux\n" + self.output.summary_simul)
            ax1.set_xscale("log")
            ax1.set_yscale("linear")

            khE = khE + 1

            if delta_t != 0.0:
                for it in range(imin_plot, imax_plot, delta_i_plot):
                    transferEKr = dset_transfer2D_EKr[it]
                    transferEAr = dset_transfer2D_EAr[it]
                    PiEKr = cumsum_inv(transferEKr) * self.oper.deltak
                    PiEAr = cumsum_inv(transferEAr) * self.oper.deltak
                    PiE = PiEKr + PiEAr
                    ax1.plot(khE, PiE, "k", linewidth=1)

                    convK = dset_convK2D[it]
                    CCK = cumsum_inv(convK) * self.oper.deltak
                    ax1.plot(khE, CCK, "y", linewidth=1)

                    convP = dset_convP2D[it]
                    CCP = cumsum_inv(convP) * self.oper.deltak
                    ax1.plot(khE, CCP, "y--", linewidth=1)

            # print(convK.sum()*self.oper.deltak,
            #       convP.sum()*self.oper.deltak,
            #       CCP[0], CCK[0])

            transferEKr = dset_transfer2D_EKr[imin_plot:imax_plot].mean(0)
            transferEKd = dset_transfer2D_EKd[imin_plot:imax_plot].mean(0)
            transferEAr = dset_transfer2D_EAr[imin_plot:imax_plot].mean(0)
            transferEAd = dset_transfer2D_EAd[imin_plot:imax_plot].mean(0)
            transferEPd = dset_transfer2D_EPd[imin_plot:imax_plot].mean(0)

            PiEKr = cumsum_inv(transferEKr) * self.oper.deltak
            PiEKd = cumsum_inv(transferEKd) * self.oper.deltak
            PiEAr = cumsum_inv(transferEAr) * self.oper.deltak
            PiEAd = cumsum_inv(transferEAd) * self.oper.deltak
            PiEPd = cumsum_inv(transferEPd) * self.oper.deltak

            PiEK = PiEKr + PiEKd
            PiEA = PiEAr + PiEAd
            PiE = PiEK + PiEA

            ax1.plot(khE, PiE, "k", linewidth=2)
            ax1.plot(khE, PiEK, "r", linewidth=2)
            ax1.plot(khE, PiEA, "b", linewidth=2)

            ax1.plot(khE, PiEKr, "r--", linewidth=2)
            ax1.plot(khE, PiEKd, "r:", linewidth=2)

            ax1.plot(khE, PiEAr, "b--", linewidth=2)
            ax1.plot(khE, PiEAd, "b:", linewidth=2)
            ax1.plot(khE, PiEPd, "c:", linewidth=1)

            convP = dset_convP2D[imin_plot:imax_plot].mean(0)
            convK = dset_convK2D[imin_plot:imax_plot].mean(0)

            CCP = cumsum_inv(convP) * self.oper.deltak
            CCK = cumsum_inv(convK) * self.oper.deltak

            ax1.plot(khE, CCP, "y--", linewidth=2)
            ax1.plot(khE, CCK, "y", linewidth=2)

            #         print(convK.sum()*self.oper.deltak,
            #               convP.sum()*self.oper.deltak,
            #               CCP[0], CCK[0],
            #               CCP[1], CCK[1]
            # )

            dset_transfer2D_Errr = h5file["transfer2D_Errr"]
            dset_transfer2D_Edrd = h5file["transfer2D_Edrd"]
            dset_transfer2D_Edrr_rrd = h5file["transfer2D_Edrr_rrd"]

            transferEdrr_rrd = dset_transfer2D_Edrr_rrd[imin_plot:imax_plot].mean(
                0
            )
            transferErrr = dset_transfer2D_Errr[imin_plot:imax_plot].mean(0)
            transferEdrd = dset_transfer2D_Edrd[imin_plot:imax_plot].mean(0)

            Pi_drr_rrd = cumsum_inv(transferEdrr_rrd) * self.oper.deltak
            Pi_rrr = cumsum_inv(transferErrr) * self.oper.deltak
            Pi_drd = cumsum_inv(transferEdrd) * self.oper.deltak

            ax1.plot(khE, Pi_drr_rrd, "m:", linewidth=1)
            ax1.plot(khE, Pi_rrr, "m--", linewidth=1)
            ax1.plot(khE, Pi_drd, "m-.", linewidth=1)

            dset_transfer2D_Eddd = h5file["transfer2D_Eddd"]
            dset_transfer2D_Erdr = h5file["transfer2D_Erdr"]
            dset_transfer2D_Eddr_rdd = h5file["transfer2D_Eddr_rdd"]
            dset_transfer2D_Eudeu = h5file["transfer2D_Eudeu"]

            transferEddr_rdd = dset_transfer2D_Eddr_rdd[imin_plot:imax_plot].mean(
                0
            )
            transferEddd = dset_transfer2D_Eddd[imin_plot:imax_plot].mean(0)
            transferErdr = dset_transfer2D_Erdr[imin_plot:imax_plot].mean(0)

            transferEudeu = dset_transfer2D_Eudeu[imin_plot:imax_plot].mean(0)

            Pi_ddr_rdd = cumsum_inv(transferEddr_rdd) * self.oper.deltak
            Pi_ddd = cumsum_inv(transferEddd) * self.oper.deltak
            Pi_rdr = cumsum_inv(transferErdr) * self.oper.deltak

            Pi_udeu = cumsum_inv(transferEudeu) * self.oper.deltak

            ax1.plot(khE, Pi_ddr_rdd, "c:", linewidth=1)
            ax1.plot(khE, Pi_ddd, "c--", linewidth=1)
            ax1.plot(khE, Pi_rdr, "c-.", linewidth=1)

            ax1.plot(khE, Pi_udeu, "g", linewidth=1)

            z_bottom_axe = 0.07
            height_axe = 0.17
            size_axe[1] = z_bottom_axe
            size_axe[3] = height_axe
            ax2 = fig.add_axes(size_axe)
            ax2.set_xlabel("$k_h$")
            ax2.set_ylabel("transfers")
            ax2.set_title("Charney PE flux")
            ax2.set_xscale("log")
            ax2.set_yscale("linear")

            if delta_t != 0.0:
                for it in range(imin_plot, imax_plot + 1, delta_i_plot):
                    transferCPE = dset_transfer2D_CPE[it]
                    PiCPE = cumsum_inv(transferCPE) * self.oper.deltak
                    ax2.plot(khE, PiCPE, "g", linewidth=1)

            transferCPE = dset_transfer2D_CPE[imin_plot:imax_plot].mean(0)
            PiCPE = cumsum_inv(transferCPE) * self.oper.deltak

        ax2.plot(khE, PiCPE, "m", linewidth=2)



"""Spatial means output (:mod:`fluidsim.solvers.ns3d.output.spatial_means`)
===========================================================================

.. autoclass:: SpatialMeansNS3D
   :members:
   :private-members:

"""

import os
import numpy as np

from fluiddyn.util import mpi

from fluidsim.base.output.spatial_means import SpatialMeansBase


class SpatialMeansNS3D(SpatialMeansBase):
    """Spatial means output."""

    def _save_one_time(self):
        tsim = self.sim.time_stepping.t
        self.t_last_save = tsim
        nrj_vx_fft, nrj_vy_fft, nrj_vz_fft = self.output.compute_energies_fft()
        energy_fft = nrj_vx_fft + nrj_vy_fft + nrj_vz_fft
        nrj_vx = self.sum_wavenumbers(nrj_vx_fft)
        nrj_vy = self.sum_wavenumbers(nrj_vy_fft)
        nrj_vz = self.sum_wavenumbers(nrj_vz_fft)
        energy = nrj_vx + nrj_vy + nrj_vz

        f_d, f_d_hypo = self.sim.compute_freq_diss()
        epsK = self.sum_wavenumbers(f_d * 2 * energy_fft)
        epsK_hypo = self.sum_wavenumbers(f_d_hypo * 2 * energy_fft)

        if self.sim.params.nu_4 > 0.0:
            f_d4 = self.params.nu_4 * self.oper.K4
            epsK4 = self.sum_wavenumbers(f_d4 * 2 * energy_fft)
            del f_d4

        if self.sim.params.nu_8 > 0.0:
            f_d8 = self.params.nu_8 * self.oper.K8
            epsK8 = self.sum_wavenumbers(f_d8 * 2 * energy_fft)
            del f_d8

        if self.sim.params.forcing.enable:
            deltat = self.sim.time_stepping.deltat
            forcing_fft = self.sim.forcing.get_forcing()

            fx_fft = forcing_fft.get_var("vx_fft")
            fy_fft = forcing_fft.get_var("vy_fft")
            fz_fft = forcing_fft.get_var("vz_fft")

            vx_fft = self.sim.state.state_spect.get_var("vx_fft")
            vy_fft = self.sim.state.state_spect.get_var("vy_fft")
            vz_fft = self.sim.state.state_spect.get_var("vz_fft")

            PK1_fft = np.ascontiguousarray(
                np.real(
                    vx_fft.conj() * fx_fft
                    + vy_fft.conj() * fy_fft
                    + vz_fft.conj() * fz_fft
                )
            )
            PK2_fft = (
                (abs(fx_fft) ** 2 + abs(fy_fft) ** 2 + abs(fz_fft) ** 2)
                * deltat
                / 2
            )

            PK1 = self.sum_wavenumbers(PK1_fft)
            PK2 = self.sum_wavenumbers(PK2_fft)

        if mpi.rank == 0:

            self.file.write(
                f"####\ntime = {tsim:11.5e}\n"
                f"E    = {energy:11.5e}\n"
                f"Ex   = {nrj_vx:11.5e} ; Ey   = {nrj_vy:11.5e} ; Ez   = {nrj_vz:11.5e}\n"
                f"epsK = {epsK:11.5e} ; epsK_hypo = {epsK_hypo:11.5e} ; "
                f"epsK_tot = {epsK + epsK_hypo:11.5e} \n"
            )

            if self.sim.params.nu_4 > 0.0:
                self.file.write(f"epsK4 = {epsK4:11.5e}\n")

            if self.sim.params.nu_8 > 0.0:
                self.file.write(f"epsK8 = {epsK8:11.5e}\n")

            if self.sim.params.forcing.enable:
                self.file.write(
                    f"PK1  = {PK1:11.5e} ; PK2       = {PK2:11.5e} ; "
                    f"PK_tot   = {PK1 + PK2:11.5e} \n"
                )

            self.file.flush()
            os.fsync(self.file.fileno())

        if self.has_to_plot and mpi.rank == 0:

            self.axe_a.plot(tsim, energy, "k.")

            # self.axe_b.plot(tsim, epsK_tot, 'k.')
            # if self.sim.params.forcing.enable:
            #     self.axe_b.plot(tsim, PK_tot, 'm.')

            if tsim - self.t_last_show >= self.period_show:
                self.t_last_show = tsim
                fig = self.axe_a.get_figure()
                fig.canvas.draw()

    def load(self):
        dict_results = {"name_solver": self.output.name_solver}

        with open(self.path_file) as file_means:
            lines = file_means.readlines()

        lines_t = []
        lines_E = []
        lines_Ex = []
        lines_PK = []
        lines_epsK = []
        lines_epsK4 = []
        lines_epsK8 = []

        for il, line in enumerate(lines):
            if line.startswith("time ="):
                lines_t.append(line)
            if line.startswith("E    ="):
                lines_E.append(line)
            if line.startswith("Ex   ="):
                lines_Ex.append(line)
            if line.startswith("PK1  ="):
                lines_PK.append(line)
            if line.startswith("epsK ="):
                lines_epsK.append(line)
            elif line.startswith("epsK4 ="):
                lines_epsK4.append(line)
            elif line.startswith("epsK8 ="):
                lines_epsK8.append(line)

        nt = len(lines_t)
        if nt > 1:
            nt -= 1

        t = np.empty(nt)
        E = np.empty(nt)
        Ex = np.empty(nt)
        Ey = np.empty(nt)
        Ez = np.empty(nt)
        PK1 = np.zeros(nt)
        PK2 = np.zeros(nt)
        PK_tot = np.zeros(nt)
        epsK = np.empty(nt)
        epsK_hypo = np.zeros(nt)
        epsK_tot = np.zeros(nt)

        if lines_epsK4:
            epsK4 = np.empty(nt)

        if lines_epsK8:
            epsK8 = np.empty(nt)

        for il in range(nt):
            line = lines_t[il]
            words = line.split()
            t[il] = float(words[2])

            line = lines_E[il]
            words = line.split()
            E[il] = float(words[2])

            line = lines_Ex[il]
            words = line.split()
            Ex[il] = float(words[2])
            Ey[il] = float(words[6])
            Ez[il] = float(words[10])

            if self.sim.params.forcing.enable:
                line = lines_PK[il]
                words = line.split()
                PK1[il] = float(words[2])
                PK2[il] = float(words[6])
                PK_tot[il] = float(words[10])

            line = lines_epsK[il]
            words = line.split()
            epsK[il] = float(words[2])
            epsK_hypo[il] = float(words[6])
            epsK_tot[il] = float(words[10])

            if lines_epsK4:
                line = lines_epsK4[il]
                words = line.split()
                epsK4[il] = float(words[2])

            if lines_epsK8:
                line = lines_epsK8[il]
                words = line.split()
                epsK8[il] = float(words[2])

        dict_results["t"] = t
        dict_results["E"] = E
        dict_results["Ex"] = Ex
        dict_results["Ey"] = Ey
        dict_results["Ez"] = Ez

        dict_results["PK1"] = PK1
        dict_results["PK2"] = PK2
        dict_results["PK_tot"] = PK_tot

        dict_results["epsK"] = epsK
        dict_results["epsK_hypo"] = epsK_hypo
        dict_results["epsK_tot"] = epsK_tot

        if lines_epsK4:
            dict_results["epsK4"] = epsK4

        if lines_epsK8:
            dict_results["epsK8"] = epsK8

        return dict_results

    def plot(self, plot_injection=True, plot_hyper=False):
        dict_results = self.load()

        t = dict_results["t"]
        E = dict_results["E"]
        Ex = dict_results["Ex"]
        Ey = dict_results["Ey"]
        Ez = dict_results["Ez"]

        epsK = dict_results["epsK"]
        epsK_hypo = dict_results["epsK_hypo"]
        epsK_tot = dict_results["epsK_tot"]

        fig, ax = self.output.figure_axe()
        ax.set_title("Energy and enstrophy\n" + self.output.summary_simul)
        ax.set_ylabel("$E(t)$")
        ax.set_xlabel("$t$")
        ax.plot(t, E, "k", linewidth=2)
        ax.plot(t, Ex, "b")
        ax.plot(t, Ey, "r")
        ax.plot(t, Ez, "c")

        fig, ax = self.output.figure_axe()
        ax.set_title(
            "Dissipation of energy and enstrophy\n" + self.output.summary_simul
        )
        ax.set_ylabel(r"$\epsilon_K(t)$")
        ax.set_xlabel("$t$")

        def _plot(x, y, fmt, label=None, linewidth=1, zorder=10):
            ax.plot(
                x,
                y,
                fmt,
                label=label,
                linewidth=linewidth,
                zorder=zorder,
            )

        _plot(t, epsK, "r", r"$\epsilon$", linewidth=2)
        if self.sim.params.nu_m4 != 0:
            _plot(t, epsK_hypo, "g", r"$\epsilon_{-4}$", linewidth=2)
            _plot(t, epsK_tot, "k", r"$\epsilon_{tot}$", linewidth=2)

        if "epsK4" in dict_results and plot_hyper:
            epsK4 = dict_results["epsK4"]
            if not np.allclose(epsK, epsK4):
                _plot(
                    t,
                    epsK4,
                    "r:",
                    r"$\epsilon_4$",
                )

        if "epsK8" in dict_results and plot_hyper:
            epsK8 = dict_results["epsK8"]
            if not np.allclose(epsK, epsK8):
                _plot(
                    t,
                    epsK8,
                    "r:",
                    r"$\epsilon_8$",
                )

        if "PK_tot" in dict_results and plot_injection:
            PK_tot = dict_results["PK_tot"]
            ax.plot(t, PK_tot, "r--", label=r"$P$", zorder=0)

        ax.legend()

    def plot_dt_E(self):
        dict_results = self.load()

        times = dict_results["t"]
        E_tot = dict_results["E"]

        dt_E_tot = np.gradient(E_tot, times)
        try:
            eps_tot = dict_results["eps_tot"]
        except KeyError:
            eps_tot = dict_results["epsK_tot"]

        all_terms = -eps_tot.copy()

        if "PK_tot" in dict_results:
            P_tot = dict_results["PK_tot"]
            if "PA_tot" in dict_results:
                P_tot += dict_results["PA_tot"]
            all_terms += P_tot

        fig, ax = self.output.figure_axe()

        ax.plot(times, dt_E_tot, "--k", label="$d_t E$")
        if "PK_tot" in dict_results:
            ax.plot(times, P_tot, "b", label="forcing")
        ax.plot(times, -eps_tot, color="orange", label="viscosity")
        ax.plot(times, all_terms, "r", label="All terms")

        ax.set_title(self.output.summary_simul)
        ax.set_xlabel("time")

        fig.legend()
        fig.tight_layout()
